"""
Implementation of Cobra using HTTP
"""

import os
import json
import time
import types
import urllib
import base64
import Cookie
import httplib
import logging
import urlparse
import traceback
import BaseHTTPServer

import cobra as c_cobra

from threading import current_thread,Thread,RLock
from SocketServer import ThreadingTCPServer

logger = logging.getLogger(__name__)
daemon = None
version = "Cobra2"
COBRA_PORT = 80
COBRASSL_PORT = 443
cobra_retrymax = None  # Optional *global* retry max count


class CobraHttpException(Exception):
    """Base for Cobra exceptions"""
    pass


def chopCobraHttpUri(uri):
    '''
    Breaks apart a URI provided by as input to a function.
    Any API call must at least contain an empty args=[] or args=()
    calls can ommit kwargs.
    URI format is:
    /[OBJECT]/[FUNCTION]?args=<JSON ARGS>&kwargs=<JSON KWARGS>

    Returns:
        [OBJECT], [FUNCTION], args or (), kwargs or {}
    '''
    url = uri
    if uri.startswith('/'):
        # get rid of starting /
        url = urlparse.urlsplit(uri[1:])

    objname, methname = url.path.split('/', 1)

    # Do we have any URL options?
    urlparams = {}
    for urlopt in urllib.unquote(url.query).split('&'):
        urlval = '1'
        if urlopt.find('=') != -1:
            urlopt, urlval = urlopt.split('=', 1)

        urlopt = urlopt.lower()
        urlparams[urlopt] = json.loads(urlval)

    return objname, methname, urlparams.get('args', ()), urlparams.get('kwargs', {})


class CobraHttpRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    '''
    Request handler for HTTP cobra
    '''
    def __init__(self, *args, **kwargs):
        self.handlers = {
                '__cobra_hello': self.handleHello,
                '__cobra_getattr': self.handleGetAttr,
                '__cobra_setattr': self.handleSetAttr,
                '__cobra_login': self.handleLogin
            }
        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def do_POST(self):
        # this is a json posted data list (args, kwargs)
        body = json.loads(urllib.unquote(self.rfile.read(int(self.headers['Content-Length']))))

        # self.handleClient(s.server, s, s.path, body=body)
        self.handleClient(body=body)

    def do_GET(self):
        self.handleClient()

    def setup(self):
        # SSLify..
        if self.server.sslkey:
            import ssl
            sslca = self.server.sslca
            keyfile = self.server.sslkey
            certfile = self.server.sslcrt
            sslreq = ssl.CERT_NONE
            # If they specify a CA key, require valid client certs
            if sslca:
                sslreq = ssl.CERT_REQUIRED

            self.request = ssl.wrap_socket(self.request,
                                           keyfile=keyfile,
                                           certfile=certfile,
                                           ca_certs=sslca,
                                           cert_reqs=sslreq,
                                           server_side=True)
        BaseHTTPServer.BaseHTTPRequestHandler.setup(self)

    def handleClient(self, body=None):
        # validate authentication
        try:
            objname, methname, args, kwargs = chopCobraHttpUri(self.path)
        except:
            # If we had an exception its a malformed request..
            self.request.send_response(httplib.INTERNAL_SERVER_ERROR)
            self.request.end_headers()
            excinfo = "%s" % traceback.format_exc()
            self.request.wfile.write(json.dumps(excinfo))
            return

        if body:
            args = body.get('args', ())
            kwargs = body.get('kwargs', {})

        # user is trying to authenticate
        if methname == '__cobra_login':
            return self.handlers['__cobra_login'](objname, kwargs)

        if self.server.authmod:
            authinfo = None
            if not self.headers.getheader('Cookie'):
                # user not authenticated
                self.request.send_response(httplib.UNAUTHORIZED)
                self.request.end_headers()
                return

            cookie = self.headers['Cookie']
            # Cookie: SessionId="[SESSIONKEY]"
            sessid = cookie.split('=',1)[1].replace("\"", "")
            # Invalid session!
            if not self.server.sessions.get(sessid):
                self.request.send_response(httplib.UNAUTHORIZED)
                self.request.end_headers()
                return

            authinfo,tstamp = self.server.sessions[sessid]

            # Ensure user cna access our object
            if not authinfo or not self.server.authmod.checkUserAccess( authinfo['user'], methname):
                self.request.send_response(httplib.UNAUTHORIZED)
                self.request.end_headers()
                return

        if self.handlers.get(methname):
            return self.handlers[methname](objname, *args, **kwargs)

        # at this point we have a call
        obj = self.server.getSharedObject(objname)
        if hasattr(obj, methname):
            meth = getattr(obj, methname)
            try:
                ret = meth(*args, **kwargs)
                self.send_response(httplib.OK)
                self.end_headers()
                self.wfile.write(json.dumps(ret))
            except Exception as e:
                self.send_response(httplib.INTERNAL_SERVER_ERROR)
                self.end_headers()
                excinfo = "%s" % traceback.format_exc()
                self.wfile.write(json.dumps(excinfo))

            return

        self.send_response(httplib.BAD_REQUEST)
        self.end_headers()
        excinfo = "Bad Request: No function named '%s'" % methname
        self.wfile.write(json.dumps(excinfo))
        return

    def handleLogin(self, objname, authinfo):
        if not self.server.getSharedObject(objname):
            raise Exception('Shared object does not exists!')

        if self.server.authmod:
            if not authinfo or not self.server.authmod.authCobraUser(authinfo):
                self.send_response(httplib.UNAUTHORIZED)
                self.end_headers()
            else:
                sesskey = base64.b64encode( os.urandom(32) )
                self.server.sessions[ sesskey ] = (authinfo, time.time())
                c = Cookie.SimpleCookie()
                c['SessionId'] = sesskey
                # set morsel
                self.send_response(httplib.OK)
                self.send_header('Set-Cookie', c.values()[0].output(header=''))
                self.send_header("Content-type", "text/html")
                self.end_headers()
            return

        self.send_response(httplib.OK)
        self.end_headers()

    def handleHello(self, objname):
        '''
        Hello messages are used to get the initial cache of
        method names for the newly connected object.
        '''
        logger.debug('Hello')
        obj = self.server.getSharedObject(objname)
        ret = {}
        for name in dir(obj):
            if type(getattr(obj, name)) in (types.MethodType, types.BuiltinMethodType):
                ret[name] = True
        self.send_response(httplib.OK)
        self.end_headers()
        self.wfile.write(json.dumps(ret))
        return

    def handleGetAttr(self, objname, attr):
        logger.debug('GetAttr: %s', attr)
        if not self.server.attr:
            self.send_response(httplib.FORBIDDEN)
            self.end_headers()
            excinfo = "__getattr__ disabled"
            self.wfile.write(json.dumps(excinfo))
            return
        obj = self.server.getSharedObject(objname)
        try:
            val = getattr(obj, attr)
            self.send_response(httplib.OK)
            self.end_headers()
            self.wfile.write(json.dumps(val))
        except Exception:
            self.send_response(httplib.NOT_FOUND)
            self.end_headers()
            excinfo = "%s" % traceback.format_exc()
            self.wfile.write(json.dumps(excinfo))

    def handleSetAttr(self, objname, name, value):
        logger.debug('SetAttr %s = %s', name, value)
        if not self.server.attr:
            self.send_response(httplib.FORBIDDEN)
            self.end_headers()
            excinfo = "__setattr__ disabled"
            self.wfile.write(json.dumps(excinfo))
            return
        obj = self.server.getSharedObject(objname)
        setattr(obj, name, value)
        self.send_response(httplib.OK)
        self.end_headers()


class CobraHttpDaemon(ThreadingTCPServer):
    def __init__(self, host="", port=COBRA_PORT, sslcrt=None, sslkey=None, sslca=None, sess_timeout=24, attr=True):
        '''
        Construct a cobra daemon object.

        Parameters:
        host        - Optional hostname/ip to bind the service to (default: inaddr_any)
        port        - The port to bind (Default: COBRA_PORT)
        timeout     - The length any session can last, before forcing reconnect
        attr        - Toggle to turn off the ability to set or get attributes

        # SSL Options
        sslcrt / sslkey     - Specify sslcrt and sslkey to enable SSL server side
        sslca               - Specify an SSL CA key to use validating client certs

        '''
        self.shared = {}
        self.host = host
        self.port = port
        self.reflock = RLock()
        self.refcnts = {}
        self.authmod = None
        self.attr    = attr

        self.sessions = {} # authenticated sessions
        self.sess_timeout=sess_timeout*60

        # SSL Options
        self.sslca = sslca
        self.sslcrt = sslcrt
        self.sslkey = sslkey

        if sslcrt and not os.path.isfile(sslcrt):
            raise Exception('CobraDaemon: sslcrt param must be a file!')

        if sslkey and not os.path.isfile(sslkey):
            raise Exception('CobraDaemon: sslkey param must be a file!')

        if sslca and not os.path.isfile(sslca):
            raise Exception('CobraDaemon: sslca param must be a file!')


        #ThreadingTCPServer.__init__(self, (host, port), CobraHttpConnectionHandler)
        ThreadingTCPServer.__init__(self, (host, port), CobraHttpRequestHandler)

        t = Thread(target=self._timeoutSessions, daemon=True)
        t.start()

        if port == 0:
            self.port = self.socket.getsockname()[1]

        self.daemon_threads = True
        self.recvtimeout = None

    def _timeoutSessions(self):
        while True:
            time.sleep( self.sess_timeout  )
            for key, (authinfo, tstamp) in self.sessions.items():
                if time.time()-tstamp > self.sess_timeout:
                    self.sessions.pop(key)

    def fireThread(self):
        thr = Thread(target=self.serve_forever, daemon=True)
        thr.start()

    def setAuthModule(self, authmod):
        '''
        Enable an authentication module for this server
        ( all connections *must* be authenticated through the authmod )

        NOTE: See cobra.auth.* for various auth module implementations

        Example:
            import cobra.auth.shadow as c_a_shadow
            authmod = c_a_shadow.ShadowFileAuth('passwdfile.txt')
            cdaemon = CobraDaemon()
            cdaemon.setAuthModule()
        '''
        self.authmod = authmod

    def getSharedObject(self, name):
        return self.shared.get(name, None)

    def shareObject(self, obj, name=None, doref=False):
        """
        Share an object in this cobra server.  By specifying
        doref=True you will let CobraProxy objects decide that
        the object is done and should be un-shared.  Also, if
        name is None a random name is chosen.

        Returns: name (or the newly generated random one)
        """
        refcnt = None
        if doref:
            refcnt = 0
        if name is None:
            raise Exception("You must specify an object name to share!")

        self.shared[name] = obj
        if doref:
            raise Exception("CobraHttp doesnt do refcnt")
        return name


class CobraHttpProxy:
    """
    A proxy object for remote objects shared with Cobra

    A few optional keyword arguments are handled by all cobra protocols:
        timeout     - Socket timeout for a cobra socket
        authinfo    - A dict, probably like {'user':'username','passwd':'mypass'}
                      ( but it can be auth module specific )
        retrymax    - Maximum number of times to try and reconnect to the HTTP server

    Additional keyword arguments may depend on protocol.

    http://
        Only the standard args

    https://
        sslca           - File path to a CA certs file.  Causes server validation.
        sslcrt / sslkey - Client side cert info

    """
    def __init__(self, URI, retrymax=None, timeout=None, **kwargs):

        scheme, host, port, name, urlparams = c_cobra.chopCobraUri(URI)
        logger.debug('Host: %s, Port: %s, Obj: %s', host, port, name)
        self._cobra_uri = URI
        self._cobra_scheme = scheme
        self._cobra_host = host
        self._cobra_port = port
        self._cobra_slookup = (host, port)
        self._cobra_name = name
        self._cobra_timeout = timeout
        self._cobra_kwargs = kwargs
        self._cobra_gothello = False
        self._cobra_sessid = None

        if kwargs.get('sslkey') and not os.path.isfile(kwargs.get('sslkey')):
            raise Exception('CobraProxy: sslkey must be a file!')

        if kwargs.get('sslcrt') and not os.path.isfile(kwargs.get('sslcrt')):
            raise Exception('CobraProxy: sslcrt must be a file!')

        if kwargs.get('sslca') and not os.path.isfile(kwargs.get('sslca')):
            raise Exception('CobraProxy: sslca must be a file!')

        csock = self._cobra_http_getsock()

        data = CobraHttpMethod(self, '__cobra_hello')()

        self._cobra_gothello = True
        self._cobra_methods = data

    def _cobra_http_getsock(self):
        thr = current_thread()
        tsocks = getattr(thr, 'cobrahttpsocks', None)
        if tsocks is None:
            tsocks = {}
            thr.cobrahttpsocks = tsocks

        sock = tsocks.get(self._cobra_slookup)
        if not sock or sock.trashed:
            # Lets build a new socket... shall we?
            sock = self._cobra_http_newsock()

            # If we have authinfo lets authenticate
            authinfo = self._cobra_kwargs.get('authinfo')
            if authinfo is not None:
                sock.authUser(authinfo)

            tsocks[self._cobra_slookup] = sock
        return sock

    def __getattr__(self, name):
        logger.debug('GetAttr: %s', name)
        if name == "__getinitargs__":
            raise AttributeError()
        # Handle methods
        if self._cobra_methods.get(name, False):
            return CobraHttpMethod(self, name)

        return CobraHttpMethod(self, '__cobra_getattr')(name)

    def __setattr__(self, name, value):
        logger.debug('SetAttr: %s = %s', name, value)
        if name.startswith('_cobra_'):
            self.__dict__[name] = value
            return

        return CobraHttpMethod(self, '__cobra_setattr')(name, value)

    def _cobra_http_newsock(self):
        """
        This is only used by *clients*
        """
        if self._cobra_scheme == 'http':
            return CobraHttpClient( CobraHttpFactory(host=self._cobra_host, port=self._cobra_port, timeout=self._cobra_timeout, factory=httplib.HTTPConnection), _cobra_name=self._cobra_name )
        else:
            return CobraHttpClient( CobraHttpFactory(host=self._cobra_host, port=self._cobra_port, timeout=self._cobra_timeout, key_file=self._cobra_kwargs.get('sslkey'), cert_file=self._cobra_kwargs.get('sslcrt'), factory=httplib.HTTPSConnection), _cobra_name=self._cobra_name )

class CobraHttpFactory:
    '''
    Cobra HTTP wrapper for HTTP connection
    '''
    def __init__(self, *args, **kwargs):
        self.http = kwargs.pop('factory')
        self.kwargs = kwargs
        self.args = args

    def __call__(self):
        return self.http(*self.args, **self.kwargs)


class CobraHttpClient:
    '''
    Wrapper for HTTP client uses Cobra web services protocol
    '''
    def __init__(self, httpfact, _cobra_name, retrymax=cobra_retrymax):
        self.httpfact = httpfact
        self.retries = 0
        self.trashed = False
        self.retrymax = retrymax
        self._cobra_name = _cobra_name
        self._cobra_sessid = None # stores our session id
        self.authinfo = None

        self.conn = httpfact()

        self.connected = True

    def authUser(self, authinfo):
        body = urllib.quote(json.dumps({'kwargs': authinfo}))
        self.conn.request("POST", "/%s/__cobra_login" %  self._cobra_name, body)

        resp = self.conn.getresponse()
        if resp.status != httplib.OK:
            raise CobraHttpException("Access Denied")

        cookie = resp.getheader('set-cookie')
        if cookie:
            self._cobra_sessid = cookie
            self.authinfo = authinfo

    def _cobra_http_geturi(self, methname, *args, **kwargs):
        uri = "/%s/%s?args=%s" % (self._cobra_name, methname, urllib.quote(json.dumps(args)))
        if kwargs:
            uri += "&kwargs=%s" % urllib.quote(json.dumps(kwargs))
        return uri

    def reConnect(self):
        while self.retrymax is None or self.retries < self.retrymax:
            logger.info('Cobra reconnection attempt')
            try:
                self.conn = self.httpfact()
                if self._cobra_sessid:
                    self.authUser(self.authinfo)
                self.retries = 0
                return
            except Exception as e:
                time.sleep(2 ** self.retries)
                self.retries += 1
        self.trashed = True

        raise CobraHttpException('Retry Exceeded!')

    def cobraHttpTransaction(self, methname, data):
        args, kwargs = data
        url = self._cobra_http_geturi(methname, *args, **kwargs)
        headers = {}

        while True:
            if self._cobra_sessid:
                headers['cookie'] = self._cobra_sessid
            self.conn.request("GET", url, headers=headers)
            resp = self.conn.getresponse()

            if resp.status == httplib.NOT_FOUND:
                exc = json.loads( resp.read())
                raise CobraHttpException(exc)

            if resp.status == httplib.UNAUTHORIZED and self._cobra_sessid:
                self.reConnect()
                continue
            if resp.status == httplib.UNAUTHORIZED:
                self.trashed = True
                raise CobraHttpException("Access Denied")
            if resp.status == httplib.INTERNAL_SERVER_ERROR:
                exc = json.loads( resp.read())
                raise CobraHttpException(exc)
            if resp.status == httplib.BAD_REQUEST:
                exc = json.loads( resp.read())
                raise CobraHttpException(exc)
            if resp.status == httplib.FORBIDDEN:
                exc = json.loads( resp.read())
                raise CobraHttpException(exc)

            data = resp.read()
            if data:
                return json.loads(data)
            return None

class CobraHttpMethod:
    def __init__(self, proxy, methname):
        self.proxy = proxy
        self.methname = methname

    def __call__(self, *args, **kwargs):
        name = self.proxy._cobra_name
        logger.debug('Calling %s:%s,(%s, %s)', name, self.methname, repr(args)[:20], repr(kwargs)[:20])
        sock = self.proxy._cobra_http_getsock()
        return sock.cobraHttpTransaction(self.methname, (args,kwargs))



"""
Initial module with supporting structures for windows
kernel serial debugging.
"""

import vstruct
from vstruct.primitives import *

# Main packet magic bytes and such
BREAKIN_PACKET                      = 0x62626262
BREAKIN_PACKET_BYTE                 = 0x62
PACKET_LEADER                       = 0x30303030
PACKET_LEADER_BYTE                  = 0x30
CONTROL_PACKET_LEADER               = 0x69696969
CONTROL_PACKET_LEADER_BYTE          = 0x69
PACKET_TRAILING_BYTE                = 0xAA

pkt_magic_names = {
    BREAKIN_PACKET:"Break Packet",
    PACKET_LEADER:"Packet",
    CONTROL_PACKET_LEADER:"Control Packet",
} 

# Primary "packet types"
PACKET_TYPE_UNUSED                  = 0
PACKET_TYPE_KD_STATE_CHANGE32       = 1
PACKET_TYPE_KD_STATE_MANIPULATE     = 2
PACKET_TYPE_KD_DEBUG_IO             = 3
PACKET_TYPE_KD_ACKNOWLEDGE          = 4
PACKET_TYPE_KD_RESEND               = 5
PACKET_TYPE_KD_RESET                = 6
PACKET_TYPE_KD_STATE_CHANGE64       = 7
PACKET_TYPE_KD_POLL_BREAKIN         = 8
PACKET_TYPE_KD_TRACE_IO             = 9
PACKET_TYPE_KD_CONTROL_REQUEST      = 10
PACKET_TYPE_KD_FILE_IO              = 11
PACKET_TYPE_MAX                     = 12

pkt_type_names = {
    PACKET_TYPE_UNUSED:"Unused",
    PACKET_TYPE_KD_STATE_CHANGE32:"State Change32",
    PACKET_TYPE_KD_STATE_MANIPULATE:"Manipulate",
    PACKET_TYPE_KD_DEBUG_IO:"Debug IO",
    PACKET_TYPE_KD_ACKNOWLEDGE:"Ack",
    PACKET_TYPE_KD_RESEND:"Resend",
    PACKET_TYPE_KD_RESET:"Reset",
    PACKET_TYPE_KD_STATE_CHANGE64:"State Change64",
    PACKET_TYPE_KD_POLL_BREAKIN:"Breakin",
    PACKET_TYPE_KD_TRACE_IO:"Trace IO",
    PACKET_TYPE_KD_CONTROL_REQUEST:"Control Request",
    PACKET_TYPE_KD_FILE_IO:"File IO",
    PACKET_TYPE_MAX:"Max",
}

# Wait State Change Types
DbgKdMinimumStateChange             = 0x00003030
DbgKdExceptionStateChange           = 0x00003030
DbgKdLoadSymbolsStateChange         = 0x00003031
DbgKdCommandStringStateChange       = 0x00003032
DbgKdMaximumStateChange             = 0x00003033

pkt_sub_wait_state_change = {
    DbgKdMinimumStateChange:"DbgKdMinimumStateChange",
    DbgKdExceptionStateChange:"DbgKdExceptionStateChange",
    DbgKdLoadSymbolsStateChange:"DbgKdLoadSymbolsStateChange",
    DbgKdCommandStringStateChange:"DbgKdCommandStringStateChange",
    DbgKdMaximumStateChange:"DbgKdMaximumStateChange",
}

# Manipulate Types
DbgKdMinimumManipulate              = 0x00003130
DbgKdReadVirtualMemoryApi           = 0x00003130
DbgKdWriteVirtualMemoryApi          = 0x00003131
DbgKdGetContextApi                  = 0x00003132
DbgKdSetContextApi                  = 0x00003133
DbgKdWriteBreakPointApi             = 0x00003134
DbgKdRestoreBreakPointApi           = 0x00003135
DbgKdContinueApi                    = 0x00003136
DbgKdReadControlSpaceApi            = 0x00003137
DbgKdWriteControlSpaceApi           = 0x00003138
DbgKdReadIoSpaceApi                 = 0x00003139
DbgKdWriteIoSpaceApi                = 0x0000313A
DbgKdRebootApi                      = 0x0000313B
DbgKdContinueApi2                   = 0x0000313C
DbgKdReadPhysicalMemoryApi          = 0x0000313D
DbgKdWritePhysicalMemoryApi         = 0x0000313E
DbgKdQuerySpecialCallsApi           = 0x0000313F
DbgKdSetSpecialCallApi              = 0x00003140
DbgKdClearSpecialCallsApi           = 0x00003141
DbgKdSetInternalBreakPointApi       = 0x00003142
DbgKdGetInternalBreakPointApi       = 0x00003143
DbgKdReadIoSpaceExtendedApi         = 0x00003144
DbgKdWriteIoSpaceExtendedApi        = 0x00003145
DbgKdGetVersionApi                  = 0x00003146
DbgKdWriteBreakPointExApi           = 0x00003147
DbgKdRestoreBreakPointExApi         = 0x00003148
DbgKdCauseBugCheckApi               = 0x00003149
DbgKdSwitchProcessor                = 0x00003150
DbgKdPageInApi                      = 0x00003151
DbgKdReadMachineSpecificRegister    = 0x00003152
DbgKdWriteMachineSpecificRegister   = 0x00003153
OldVlm1                             = 0x00003154
OldVlm2                             = 0x00003155
DbgKdSearchMemoryApi                = 0x00003156
DbgKdGetBusDataApi                  = 0x00003157
DbgKdSetBusDataApi                  = 0x00003158
DbgKdCheckLowMemoryApi              = 0x00003159
DbgKdClearAllInternalBreakpointsApi = 0x0000315A
DbgKdFillMemoryApi                  = 0x0000315B
DbgKdQueryMemoryApi                 = 0x0000315C
DbgKdSwitchPartition                = 0x0000315D
DbgKdMaximumManipulate              = 0x0000315E

pkt_sub_manipulate = {
    DbgKdMinimumManipulate:"DbgKdMinimumManipulate",
    DbgKdReadVirtualMemoryApi:"DbgKdReadVirtualMemoryApi",
    DbgKdWriteVirtualMemoryApi:"DbgKdWriteVirtualMemoryApi",
    DbgKdGetContextApi:"DbgKdGetContextApi",
    DbgKdSetContextApi:"DbgKdSetContextApi",
    DbgKdWriteBreakPointApi:"DbgKdWriteBreakPointApi",
    DbgKdRestoreBreakPointApi:"DbgKdRestoreBreakPointApi",
    DbgKdContinueApi:"DbgKdContinueApi",
    DbgKdReadControlSpaceApi:"DbgKdReadControlSpaceApi",
    DbgKdWriteControlSpaceApi:"DbgKdWriteControlSpaceApi",
    DbgKdReadIoSpaceApi:"DbgKdReadIoSpaceApi",
    DbgKdWriteIoSpaceApi:"DbgKdWriteIoSpaceApi",
    DbgKdRebootApi:"DbgKdRebootApi",
    DbgKdContinueApi2:"DbgKdContinueApi2",
    DbgKdReadPhysicalMemoryApi:"DbgKdReadPhysicalMemoryApi",
    DbgKdWritePhysicalMemoryApi:"DbgKdWritePhysicalMemoryApi",
    DbgKdQuerySpecialCallsApi:"DbgKdQuerySpecialCallsApi",
    DbgKdSetSpecialCallApi:"DbgKdSetSpecialCallApi",
    DbgKdClearSpecialCallsApi:"DbgKdClearSpecialCallsApi",
    DbgKdSetInternalBreakPointApi:"DbgKdSetInternalBreakPointApi",
    DbgKdGetInternalBreakPointApi:"DbgKdGetInternalBreakPointApi",
    DbgKdReadIoSpaceExtendedApi:"DbgKdReadIoSpaceExtendedApi",
    DbgKdWriteIoSpaceExtendedApi:"DbgKdWriteIoSpaceExtendedApi",
    DbgKdGetVersionApi:"DbgKdGetVersionApi",
    DbgKdWriteBreakPointExApi:"DbgKdWriteBreakPointExApi",
    DbgKdRestoreBreakPointExApi:"DbgKdRestoreBreakPointExApi",
    DbgKdCauseBugCheckApi:"DbgKdCauseBugCheckApi",
    DbgKdSwitchProcessor:"DbgKdSwitchProcessor",
    DbgKdPageInApi:"DbgKdPageInApi",
    DbgKdReadMachineSpecificRegister:"DbgKdReadMachineSpecificRegister",
    DbgKdWriteMachineSpecificRegister:"DbgKdWriteMachineSpecificRegister",
    OldVlm1:"OldVlm1",
    OldVlm2:"OldVlm2",
    DbgKdSearchMemoryApi:"DbgKdSearchMemoryApi",
    DbgKdGetBusDataApi:"DbgKdGetBusDataApi",
    DbgKdSetBusDataApi:"DbgKdSetBusDataApi",
    DbgKdCheckLowMemoryApi:"DbgKdCheckLowMemoryApi",
    DbgKdClearAllInternalBreakpointsApi:"DbgKdClearAllInternalBreakpointsApi",
    DbgKdFillMemoryApi:"DbgKdFillMemoryApi",
    DbgKdQueryMemoryApi:"DbgKdQueryMemoryApi",
    DbgKdSwitchPartition:"DbgKdSwitchPartition",
    DbgKdMaximumManipulate:"DbgKdMaximumManipulate",
}

# Debug I/O Types
DbgKdPrintStringApi                 = 0x00003230
DbgKdGetStringApi                   = 0x00003231

# Control Report Flags
REPORT_INCLUDES_SEGS                = 0x0001
REPORT_INCLUDES_CS                  = 0x0002

# Protocol Versions
DBGKD_64BIT_PROTOCOL_VERSION1       = 5
DBGKD_64BIT_PROTOCOL_VERSION2       = 6

# Query Memory Address Spaces
DBGKD_QUERY_MEMORY_VIRTUAL          = 0
DBGKD_QUERY_MEMORY_PROCESS          = 0
DBGKD_QUERY_MEMORY_SESSION          = 1
DBGKD_QUERY_MEMORY_KERNEL           = 2

# Query Memory Flags
DBGKD_QUERY_MEMORY_READ             = 0x01
DBGKD_QUERY_MEMORY_WRITE            = 0x02
DBGKD_QUERY_MEMORY_EXECUTE          = 0x04
DBGKD_QUERY_MEMORY_FIXED            = 0x08

ULONG = v_uint32
ULONG64 = v_uint64
BOOLEAN = v_uint32

class DBGKD_LOAD_SYMBOLS64(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self._vs_field_align = True
        self.PathNameLength = v_uint32()
        self.BaseOfDll = v_uint64()
        self.ProcessId = v_uint64()
        self.CheckSum = v_uint32()
        self.SizeOfImage = v_uint32()
        #self.UnloadSymbols = v_uint8()
        self.UnloadSymbols = v_uint32() # HACK must be 32 bit aligned

class DBGKD_WAIT_STATE_CHANGE64(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self._vs_field_align = True
        self.NewState = v_uint32()
        self.ProcessorLevel = v_uint16()
        self.Processor = v_uint16()
        self.NumberProcessors = v_uint32()
        self.Thread = v_uint64()
        self.ProgramCounter = v_uint64()




'''
If a "function" is in the plt it's a wrapper for something in the GOT.
Make that apparent.

This analysis module attempts to be platform- and architecture-agnostic.
Different conpilers make that more or less difficult.  We're jumping over
numerous hurdles to get accuracy everywhere without having different
analysis modules for each achitecture and compiler.

First of all, we start out by attempting to analyze all PLT sections
(.plt and .plt.got) before any other code is run.  This requires us to
attempt to lay down Code locations all the way through the section.
Some architectures place pointers in the section which are referenced
by code in earlier functions, as a "literal pool."

We start off analysis by identifying all the Unconditional Branches which
don't point to the beginning of the section.  Lazy loading typically ends
up with some tricks of code places a value for the function in a given
register, and branching to a trampoline function at the beginning of the
section, which then branches to the LIBC dynamic loader.  The other
Unconditional Branches should jump into the actual function, pointed to
by the GOT entry for each function.
Oh, by the way, not all PLT functions are for external Imports.  Some
compile flags wrap local functions with a PLT/GOT pair as well.

ANYWAY!  We start of identifying all the Unconditional Branches, and flagging
each with whether they are "Real PLT" branches or some Lazy Loader Trampoline
thing.
We then attempt to determine how large each PLT function is.  This means, how
far before the Branch is the start of the PLT entry.  The reason is, we know
precisely where the Branches are.  We *don't* know (before analyzing the rest
of the codebase) what will actually be called.

In order to identify the start of the function, we determine if there is a
Trampoline, and if so, skip to the next "RealPLT" branch.  Then we work
backward until we run into another unconditional branch, a NOP, a non-location,
or the start of the section.  That helps us determine where the functions all
start in relation to the branch, giving us "plt_size".

We use a heuristic calculating how many distances there are, and choose
the most common distance to determine the plt_distance, or the distance between
the start (and the branch) of each PLT entry.

Armed with plt_distance and plt_size, we then determine where each PLT entry
begins, and make those Functions, and analysis continues from there.
'''
import logging
import collections

import envi
import vivisect
import envi.common as e_cmn
import envi.archs.i386 as e_i386

logger = logging.getLogger(__name__)

MAGIC_PLT_SIZE = 16


def analyze(vw):
    """
    Do simple linear disassembly of the .plt section if present.
    Make functions
    """
    if vw.getMeta('Platform') == 'qnx':
        logger.warning('skipping initial PLT section analysis for QNX binary.')
        return

    for sva, ssize in getPLTs(vw):
        analyzePLT(vw, sva, ssize)


def getGOTByFilename(vw, filename):
    '''
    Returns GOT location for the filename specified.

    First checks through Sections (vw.getSegments), then through Dynamics.
    If the two clash, Dynamics wins.

    Linux binaries tend to have both Sections and Dynamics entries
    OpenBSD is only Sections.
    QNX has only Dynamics.
    '''
    gottup = vw.getFileMeta(filename, 'GOT')
    if gottup is not None:
        return gottup

    gotva = None
    gotsize = None
    for va, size, name, fname in vw.getSegments():
        if name == '.got.plt' and fname == filename:
            gotva = va
            gotsize = size

        elif name == ".got" and fname == filename and gotva is None:
            # for some reason, if '.got.plt' is present, it's the "right one"
            gotva = va
            gotsize = size

    # pull GOT info from Dynamics
    fdyns = vw.getFileMeta(filename, 'ELF_DYNAMICS')
    if fdyns is not None:
        FGOT = fdyns.get('DT_PLTGOT')
        if FGOT is not None:
            if FGOT != gotva and None not in (FGOT, gotva):
                logger.warning("Dynamics and Sections have different GOT entries: S:0x%x D:0x%x. using Dynamics", gotva, FGOT)

            # since Dynamics don't store the GOT size, just use to the end of the memory map
            mmva, mmsz, mmperm, mmname = vw.getMemoryMap(FGOT)
            gotva = FGOT
            moffset = gotva - mmva
            gotsize = mmsz - moffset

    vw.setFileMeta(filename, 'GOT', (gotva, gotsize))
    return gotva, gotsize

def getGOTs(vw):
    out = collections.defaultdict(set)

    gotva = None
    gotsize = None
    for va, size, name, fname in vw.getSegments():
        if name in ('.got.plt', '.got'):
            out[fname].add((va, size))

    # pull GOT info from Dynamics
    for filename in vw.getFiles():
        fdyns = vw.getFileMeta(filename, 'ELF_DYNAMICS')
        if fdyns is not None:
            FGOT = fdyns.get('DT_PLTGOT')
            if FGOT is not None:
                flist = out[filename]

                skip = False
                for va, size in flist:
                    if FGOT == va:
                        skip = True

                if skip:
                    continue

                # since Dynamics don't store the GOT size, just use to the end of the memory map
                mmva, mmsz, mmperm, mmname = vw.getMemoryMap(FGOT)
                gotva = FGOT
                moffset = gotva - mmva
                gotsize = mmsz - moffset
                flist.add((gotva, gotsize))
    return out

def getPLTs(vw):
    plts = []
    # Thought:  This is DT_PLTGOT, although each ELF will/may have their own DT_PLTGOT.
    for va, size, name, fname in vw.getSegments():
        if name.startswith(".plt") or name == '.rela.plt':
            plts.append((va, size))

    # pull GOT info from Dynamics
    for fname in vw.getFiles():
        fdyns = vw.getFileMeta(fname, 'ELF_DYNAMICS')
        if fdyns is not None:
            FPLT = fdyns.get('DT_JMPREL')
            FPLTSZ = fdyns.get('DT_PLTRELSZ')

            # if we don't have FPLT or FPLTSZ, skip this
            if FPLT is None or FPLTSZ is None:
                continue

            if vw.getFileMeta(fname, 'addbase'):
                imgbase = vw.getFileMeta(fname, 'imagebase')
                logger.debug('Adding Imagebase: 0x%x', imgbase)
                FPLT += imgbase

            newish = True
            for pltva, pltsize in plts:
                if FPLT == pltva:
                    newish = False

            if newish:
                plts.append((FPLT, FPLTSZ))

    return plts


def analyzePLT(vw, ssva, ssize):
    try:
        '''
        analyze an entire section designated as "PLT" or "PLTGOT"
        '''
        emu = None
        sva = ssva
        nextseg = sva + ssize
        gotva, gotsize = getGOTByFilename(vw, vw.getFileByVa(ssva))

        ###### make code for every opcode in PLT
        # make and parse opcodes.  keep track of unconditional branches
        branchvas = []

        # drag an emulator along to calculate branches, if available
        try:
            emu = vw.getEmulator(va=ssva)
            emu.setRegister(e_i386.REG_EBX, gotva)  # every emulator will have a 4th register, and if it's not used, no harm done.
        except Exception as e:
            logger.debug("no emulator available: %r", e, exc_info=1)
            return

        while sva < nextseg:
            logger.debug('analyzePLT(0x%x, 0x%x) first pass:  sva: 0x%x   nextseg: 0x%x', ssva, ssize, sva, nextseg)
            if vw.getLocation(sva) is None:
                logger.debug('making code: 0x%x', sva)
                try:
                    vw.makeCode(sva)
                except Exception:
                    logger.exception('0x%x: exception', sva)

            ltup = vw.getLocation(sva)

            # ltup should *always* be a location, unless something broke.  however, fail gracefully :)
            if ltup is not None:
                lva, lsz, ltype, ltinfo = ltup
                # if the location is an Opcode, check for branch info
                if ltype == vivisect.LOC_OP:
                    op = vw.parseOpcode(lva)
                    emu.setProgramCounter(lva)
                    if op.iflags & envi.IF_BRANCH and \
                            not op.iflags & envi.IF_COND:

                        # we grab all unconditional branches, and tag them.
                        # some compilers pepper conditional branches between
                        # PLT entries, for lazy-loading tricks.  We skip those.

                        # quickly, check the branch target is not the section
                        # start address.
                        tbrva = op.opers[-1].getOperValue(op, emu=emu)
                        tbrref = op.opers[-1].getOperAddr(op, emu=emu)
                        realplt = not bool(tbrva == ssva)

                        # since the above check will "fail open", refine check if True
                        if realplt:
                            loc = vw.getLocation(tbrref)
                            tgt_seg = vw.getSegment(tbrref)

                            # if the location referenced by this address doesn't exist,
                            # it's not an IMPORT
                            if loc is None:
                                realplt = False

                            # if target is in the ELF Section with a name starting with ".got"
                            # OR that the target address is after the DT_PLTGOT entry (
                            elif tgt_seg is not None:
                                tsegva, tsegsz, tsegname, tsegfile = tgt_seg
                                # see if the target address segment is GOT, but
                                # also don't let it be the first entry, which is the LazyLoader
                                if not (tsegname.startswith('.got') and tsegva != lva):
                                    logger.debug("0x%x: tbrref not in GOT (segment)", tbrref)
                                    realplt = False

                            else:
                                taintaddr = emu.getVivTaint(tbrref)
                                taintval = emu.getVivTaint(tbrva)

                                if taintaddr is not None:
                                    taintva, ttype, loctup = taintaddr
                                    if ttype != 'import':
                                        logger.debug("0x%x: is not an import taint! (%r, %r)", taintva, ttype, loctup)
                                        realplt = False

                                elif not vw.isValidPointer(tbrref):
                                    logger.debug("0x%x: tbrref is not valid pointer!", tbrref)
                                    realplt = False

                                else:
                                    fname = vw.getFileByVa(tbrref)
                                    if fname is not None:
                                        fdyns = vw.getFileMeta(fname, 'ELF_DYNAMICS')
                                        if fdyns is not None:
                                            FGOT = fdyns.get('DT_PLTGOT')
                                            if FGOT is not None and tbrref <= FGOT:
                                                logger.debug("0x%x: tbrref not in GOT (DT_PLTGOT)", tbrref)
                                                realplt = False

                        branchvas.append((op.va, realplt, tbrva, op))

                    # after analyzing the situation, emulate the opcode
                    emu.executeOpcode(op)

                sva += lsz
                logger.debug('incrementing to next va: 0x%x', sva)
            else:
                logger.warning('makeCode(0x%x) failed to make a location (probably failed instruction decode)!  incrementing instruction pointer by 1 to continue PLT analysis <fingers crossed>', sva)
                sva += 1

    except Exception as e:
        logger.exception('analyzePLT(0x%x, %r): %s', ssva, ssize, str(e))


MAX_OPS = 10

def analyzeFunction(vw, funcva):
    # check to make sure we're in the PLT
    plts = getPLTs(vw)
    isplt = False
    for pltva, pltsz in plts:
        if pltva <= funcva < (pltva + pltsz):
            isplt = True
            segva = pltva
            segsize = pltsz
            break

    # if we're not
    if not isplt:
        logger.log(e_cmn.SHITE, '0x%x: not part of a .plt section', funcva)
        return

    logger.info('analyzing PLT function: 0x%x', funcva)
    # start off spinning up an emulator to track through the PLT entry
    # slight hack, but we don't currently know if thunk_bx exists
    fname = vw.getFileByVa(funcva)
    gotva, gotsize = getGOTByFilename(vw, fname)

    # all architectures should at least have some minimal emulator
    emu = vw.getEmulator(va=funcva)
    emu.setRegister(e_i386.REG_EBX, gotva)  # every emulator will have a 4th register, and if it's not used, no harm done.

    # roll through instructions looking for a branch (pretty quickly)
    count = 0
    opva = funcva
    try:
        op = vw.parseOpcode(opva)
        while count < MAX_OPS and op.iflags & envi.IF_BRANCH == 0:
            emu.executeOpcode(op)
            opva += len(op)
            op = vw.parseOpcode(opva)
            logger.log(e_cmn.SHITE, "0x%x: %r", opva, op)
            count += 1
    except Exception as e:
        logger.warning('failure analyzing PLT func 0x%x: %r', funcva, e)
        return

    # we've reached the branch instruction or we've gone through too many opcodes
    if op.iflags & envi.IF_BRANCH == 0:
        logger.warning("PLT: 0x%x - Could not find a branch!", funcva)
        return

    # cycle through the opcode's "branches" to make connections and look for funcname.
    # BR_DEREF - the address in opval is a pointer, not code
    # BR_FALL  - the address is the next instruction (should never happen here)
    funcname = None
    branches = op.getBranches(emu)
    logger.debug("getBranches returned: %r", branches)

    for opval, brflags in branches:
        if opval is None:
            logger.info("getBranches():  opval is None!:  op = %r     branches = %r", op, branches)
        else:
            logger.debug('getBranches():  ref: 0x%x  brflags: 0x%x', opval, brflags)

        loctup = vw.getLocation(opval)

        # if BR_DEREF, this is a pointer
        if brflags & envi.BR_DEREF:
            if loctup is None:              ###### TODO: CHECK HERE FOR TAINT VALUE!
                taintval = emu.getVivTaint(opval)
                if taintval:
                    taintrepr = emu.reprVivTaint(taintval)
                else:
                    taintrepr = 'None'
                logger.error('0x%x: opval=0x%x: brflags is BR_DEREF, but loctup is None.  Don\'t know what to do. skipping.  taint:%r (%r)', op.va, opval, taintval, taintrepr)
                continue

            lva, lsz, ltype, ltinfo = loctup
            opref = opval
            # add the xref to whatever location referenced (assuming the opref hack worked)
            if vw.isValidPointer(opref):
                logger.debug('reference 0x%x is valid, adding Xref', opref)
                vw.addXref(op.va, opref, vivisect.REF_DATA)

            if ltype == vivisect.LOC_IMPORT:
                # import locations store the name as ltinfo
                funcname = ltinfo
                logger.debug("0x%x: (0x%x) LOC_IMPORT by BR_DEREF %r", funcva, opval, funcname)

            elif ltype == vivisect.LOC_POINTER:
                # we have a deref to a pointer.
                funcname = vw.getName(ltinfo)
                if ltinfo:
                    logger.debug("0x%x: (0x%x->0x%x) LOC_POINTER by BR_DEREF %r", funcva, opval, ltinfo, funcname)
                else:
                    logger.debug("0x%x: (0x%x->%r) LOC_POINTER by BR_DEREF %r", funcva, opval, ltinfo, funcname)
            else:
                logger.warning("0x%x: (0x%x) not LOC_IMPORT or LOC_POINTER?? by BR_DEREF %r", funcva, opval, loctup)

        else:
            # check the taint tracker to determine if it's an import (the opval value is pointless if it is)
            taint = emu.getVivTaint(opval)
            if taint is not None:
                # if it is an import taint
                taintva, ttype, loctup = taint
                if ttype != 'import':
                    logger.warning('getBranches(): returned a Taint which is *not* an import: %r', taint)
                    return

                lva, lsz, ltype, ltinfo = loctup
                funcname = ltinfo
                logger.debug('0x%x: LOC_IMPORT (emu-taint): 0x%x:  %r', opva, lva, funcname)
                if not vw.getXrefsFrom(opva):
                    vw.addXref(opva, lva, vivisect.REF_DATA)

            else:
                # instead of a taint (which *should* indicate an IMPORT), we have real pointer.
                # check the location type
                if not vw.isValidPointer(opval):
                    logger.info("PLT: opval=0x%x - not a valid pointer... skipping", opval)
                    continue

                if loctup is None:
                    if opval is None:
                        logger.warning("PLT: 0x%x - branch deref not defined: (opval is None!)", opva)
                        return
                    else:
                        aopval, aflags = vw.arch.archModifyFuncAddr(opval, {'arch': envi.ARCH_DEFAULT})
                        logger.info("PLT: 0x%x - making function at location 0x%x", opva, aopval)
                        vw.makeFunction(aopval, arch=aflags['arch'])
                        loctup = vw.getLocation(opval)

                lva, lsz, ltype, ltinfo = loctup

                # in case the architecture cares about the function address...
                aopval, aflags = vw.arch.archModifyFuncAddr(opval, {'arch': envi.ARCH_DEFAULT})
                funcname = vw.getName(aopval)
                if funcname is None:
                    funcname = vw.getName(opval)

                # sort through the location types and adjust accordingly
                if ltype == vivisect.LOC_IMPORT:
                    logger.info("0x%x: (0x%x) dest is LOC_IMPORT but missed taint for %r", funcva, opval, funcname)
                    # import locations store the name as ltinfo
                    funcname = ltinfo

                elif ltype == vivisect.LOC_OP:
                    logger.debug("0x%x: succeeded finding LOC_OP at the end of the rainbow! (%r)", funcva, funcname)
                    if vw.getFunction(aopval) is None:
                        logger.debug("0x%x: code does not exist at 0x%x.  calling makeFunction()", funcva, aopval)
                        vw.makeFunction(aopval, arch=aflags['arch'])

                    # this "thunk" actually calls something in the workspace, that exists as a function...
                    logger.info('0x%x points to real code (0x%x: %r)', funcva, opval, funcname)
                    vw.addXref(op.va, aopval, vivisect.REF_CODE)
                    vw.setVaSetRow('FuncWrappers', (funcva, opval))

                elif ltype == vivisect.LOC_POINTER:
                    logger.info("0x%x: (0x%x) dest is LOC_POINTER -> 0x%x", funcva, opval, ltinfo)
                    funcname = ltinfo


            if vw.getFunction(opval) == segva:
                # this is a lazy-link/load function, calling the first entry in the PLT
                logger.debug('skipping lazy-loader function: 0x%x (calls 0x%x)', funcva, opval)
                return

    # now make the thunk with appropriate naming
    # if we can't resolve a name, don't make it a thunk
    if funcname is None:
        logger.warning('0x%x: FAIL: could not resolve name for 0x%x.  Skipping.', funcva, opval)
        return

    # trim up the name
    if funcname.startswith('*.'):
        funcname = funcname[2:]

    # if filelocal for the target symbol, strip off the filename
    if funcname.startswith(fname + "."):
        funcname = funcname[len(fname)+1:]

    # trim off any address at the end
    if funcname.endswith('_%.8x' % opval):
        funcname = funcname[:-9]

    logger.info('makeFunctionThunk(0x%x, "plt_%s")', funcva, funcname)
    vw.makeFunctionThunk(funcva, "*." + funcname, addVa=False, filelocal=True,
            basename="plt_" + funcname)

'''
def printPLTs(vw):
    for seg in vw.getSegments():
        if seg[2] in ('.plt', '.plt.got'):
            print seg[2]
            vwdis(vw, seg[0], 10)

linux/i386/vdir:
.plt
0x80494c0:                        ff3504100608  push dword [0x08061004]
0x80494c6:                        ff2508100608  jmp dword [0x08061008]
0x80494cc:                                0000  add byte [eax],al
0x80494ce:                                0000  add byte [eax],al
0x80494d0:                        ff250c100608  jmp dword [0x0806100c]
0x80494d6:                          6800000000  push 0
0x80494db:                          e9e0ffffff  jmp 0x080494c0
0x80494e0:                        ff2510100608  jmp dword [0x08061010]
0x80494e6:                          6808000000  push 8
0x80494eb:                          e9d0ffffff  jmp 0x080494c0
.plt.got
0x8049b60:                        ff25fc0f0608  jmp dword [0x08060ffc]
0x8049b66:                                6690  nop
0x8049b68:                                0000  add byte [eax],al
0x8049b6a:                                0000  add byte [eax],al
0x8049b6c:                                0000  add byte [eax],al
0x8049b6e:                                0000  add byte [eax],al

linux/ie86/chgrp:
.plt
0x8048d90:                        ff3504400508  push dword [0x08054004]
0x8048d96:                        ff2508400508  jmp dword [0x08054008]
0x8048d9c:                                0000  add byte [eax],al
0x8048d9e:                                0000  add byte [eax],al
0x8048da0:                        ff250c400508  jmp dword [0x0805400c]
0x8048da6:                          6800000000  push 0
0x8048dab:                          e9e0ffffff  jmp 0x08048d90
0x8048db0:                        ff2510400508  jmp dword [0x08054010]
0x8048db6:                          6808000000  push 8
0x8048dbb:                          e9d0ffffff  jmp 0x08048d90
.plt.got
0x80491d0:                        ff25fc3f0508  jmp dword [0x08053ffc]
0x80491d6:                                6690  nop
0x80491d8:                                0000  add byte [eax],al
0x80491da:                                0000  add byte [eax],al


linux/i386/libc-2.13.so:
.plt
0x2016b78:                        ffb304000000  push dword [ebx + 4]
0x2016b7e:                        ffa308000000  jmp dword [ebx + 8]
0x2016b84:                                0000  add byte [eax],al
0x2016b86:                                0000  add byte [eax],al
0x2016b88:                        ffa30c000000  jmp dword [ebx + 12]
0x2016b8e:                          6800000000  push 0
0x2016b93:                          e9e0ffffff  jmp 0x02016b78
0x2016b98:                        ffa310000000  jmp dword [ebx + 16]
0x2016b9e:                          6808000000  push 8
0x2016ba3:                          e9d0ffffff  jmp 0x02016b78


linux/i386/libstdc++.so.6.0.25.viv
.plt
0x206aeb0:                        ffb304000000  push dword [ebx + 4]
0x206aeb6:                        ffa308000000  jmp dword [ebx + 8]
0x206aebc:                                0000  add byte [eax],al
0x206aebe:                                0000  add byte [eax],al
0x206aec0:                        ffa30c000000  jmp dword [ebx + 12]
0x206aec6:                          6800000000  push 0
0x206aecb:                          e9e0ffffff  jmp 0x0206aeb0
0x206aed0:                        ffa310000000  jmp dword [ebx + 16]
0x206aed6:                          6808000000  push 8
0x206aedb:                          e9d0ffffff  jmp 0x0206aeb0
.plt.got
0x206e2e0:                        ffa3b4f7ffff  jmp dword [ebx - 2124]
0x206e2e6:                                6690  nop
0x206e2e8:                        ffa3c8f7ffff  jmp dword [ebx - 2104]
0x206e2ee:                                6690  nop
0x206e2f0:                        ffa334f8ffff  jmp dword [ebx - 1996]
0x206e2f6:                                6690  nop
0x206e2f8:                        ffa34cf8ffff  jmp dword [ebx - 1972]
0x206e2fe:                                6690  nop
0x206e300:                        ffa35cf8ffff  jmp dword [ebx - 1956]
0x206e306:                                6690  nop

linux/libstdc++.so.6.viv
.plt
0x2058cc0:                        ff3542432900  push qword [rip + 2704194]
0x2058cc6:                        ff2544432900  jmp qword [rip + 2704196]
0x2058ccc:                            0f1f4000  nop dword [rax]
0x2058cd0:                        ff2542432900  jmp qword [rip + 2704194]
0x2058cd6:                          6800000000  push 0
0x2058cdb:                          e9e0ffffff  jmp 0x02058cc0
0x2058ce0:                        ff253a432900  jmp qword [rip + 2704186]
0x2058ce6:                          6801000000  push 1
0x2058ceb:                          e9d0ffffff  jmp 0x02058cc0
0x2058cf0:                        ff2532432900  jmp qword [rip + 2704178]

linux/i386/static32.llvm.elf.viv
.plt
0x80481c8:                        ff250cb00d08  jmp dword [0x080db00c]
0x80481ce:                                6690  nop
0x80481d0:                        ff2510b00d08  jmp dword [0x080db010]
0x80481d6:                                6690  nop
0x80481d8:                        ff2514b00d08  jmp dword [0x080db014]
0x80481de:                                6690  nop
0x80481e0:                        ff2518b00d08  jmp dword [0x080db018]
0x80481e6:                                6690  nop
0x80481e8:                        ff251cb00d08  jmp dword [0x080db01c]
0x80481ee:                                6690  nop

linux/arm/sh.viv
.plt
0xb5e4:                       04e02de5  push lr
0xb5e8:                       04e09fe5  ldr lr, [#0xb5f4]
0xb5ec:                       0ee08fe0  add lr, pc, lr
0xb5f0:                       08f0bee5  ldr pc, [lr, #0x8]!
0xb5f4:                       9cde0300  muleq r3, r12, lr
0xb5f8:                       00c68fe2  adr r12, 0x0000b600
0xb5fc:                       3dca8ce2  add r12, r12, #0x3d000
0xb600:                       9cfebce5  ldr pc, [r12, #0xe9c]!
0xb604:                       00c68fe2  adr r12, 0x0000b60c
0xb608:                       3dca8ce2  add r12, r12, #0x3d000

linux/amd64/ls.viv
.plt
0x401ff0:                         ff35fa6f2100  push qword [rip + 2191354]
0x401ff6:                         ff25fc6f2100  jmp qword [rip + 2191356]
0x401ffc:                             0f1f4000  nop dword [rax]
0x402000:                         ff25fa6f2100  jmp qword [rip + 2191354]
0x402006:                           6800000000  push 0
0x40200b:                           e9e0ffffff  jmp 0x00401ff0
0x402010:                         ff25f26f2100  jmp qword [rip + 2191346]
0x402016:                           6801000000  push 1
0x40201b:                           e9d0ffffff  jmp 0x00401ff0
0x402020:                         ff25ea6f2100  jmp qword [rip + 2191338]

linux/amd64/libc-2.27-oldstyle.so
.plt
0x2020fd0:                        ff3532a03c00  push qword [rip + 3973170]
0x2020fd6:                        ff2534a03c00  jmp qword [rip + 3973172]
0x2020fdc:                            0f1f4000  nop dword [rax]
0x2020fe0:                        ff2532a03c00  jmp qword [rip + 3973170]
0x2020fe6:                          682d000000  push 45
0x2020feb:                          e9e0ffffff  jmp 0x02020fd0
0x2020ff0:                        ff252aa03c00  jmp qword [rip + 3973162]
0x2020ff6:                          682c000000  push 44
0x2020ffb:                          e9d0ffffff  jmp 0x02020fd0
0x2021000:                        ff2522a03c00  jmp qword [rip + 3973154]
.plt.got
0x20212c0:                        ff256a9b3c00  jmp qword [rip + 3971946]
0x20212c6:                                6690  nop
0x20212c8:                        ff25ca9c3c00  jmp qword [rip + 3972298]
0x20212ce:                                6690  nop
0x20212d0:                                  55  push rbp
0x20212d1:                                  53  push rbx
0x20212d2:                      488d3d0d251900  lea rdi,qword [rip + 1647885]
0x20212d9:                          be02000080  mov esi,0x80000002
0x20212de:                            4883ec08  sub rsp,8
0x20212e2:                          e869511400  call 0x02166450

linux/amd64/libstdc++.so.6.0.25.viv
.plt
0x20884a0:                        ff35629b2f00  push qword [rip + 3119970]
0x20884a6:                        ff25649b2f00  jmp qword [rip + 3119972]
0x20884ac:                            0f1f4000  nop dword [rax]
0x20884b0:                        ff25629b2f00  jmp qword [rip + 3119970]
0x20884b6:                          6800000000  push 0
0x20884bb:                          e9e0ffffff  jmp 0x020884a0
0x20884c0:                        ff255a9b2f00  jmp qword [rip + 3119962]
0x20884c6:                          6801000000  push 1
0x20884cb:                          e9d0ffffff  jmp 0x020884a0
0x20884d0:                        ff25529b2f00  jmp qword [rip + 3119954]
.plt.got
0x208b800:                        ff2542582f00  jmp qword [rip + 3102786]
0x208b806:                                6690  nop
0x208b808:                        ff254a582f00  jmp qword [rip + 3102794]
0x208b80e:                                6690  nop
0x208b810:                        ff2512592f00  jmp qword [rip + 3102994]
0x208b816:                                6690  nop
0x208b818:                        ff253a592f00  jmp qword [rip + 3103034]
0x208b81e:                                6690  nop
0x208b820:                        ff255a592f00  jmp qword [rip + 3103066]
0x208b826:                                6690  nop

linux/amd64/static64.llvm.elf.viv
.plt
0x400428:                         ff25eaab2b00  jmp qword [rip + 2862058]
0x40042e:                                 6690  nop
0x400430:                         ff25eaab2b00  jmp qword [rip + 2862058]
0x400436:                                 6690  nop
0x400438:                         ff25eaab2b00  jmp qword [rip + 2862058]
0x40043e:                                 6690  nop
0x400440:                         ff25eaab2b00  jmp qword [rip + 2862058]
0x400446:                                 6690  nop
0x400448:                         ff25eaab2b00  jmp qword [rip + 2862058]
0x40044e:                                 6690  nop

linux/amd64/libc-2.27.so.viv
.plt
0x2020fd0:                        ff3532a03c00  push qword [rip + 3973170]
0x2020fd6:                        ff2534a03c00  jmp qword [rip + 3973172]
0x2020fdc:                            0f1f4000  nop dword [rax]
0x2020fe0:                        ff2532a03c00  jmp qword [rip + 3973170]
0x2020fe6:                          682d000000  push 45
0x2020feb:                          e9e0ffffff  jmp 0x02020fd0
0x2020ff0:                        ff252aa03c00  jmp qword [rip + 3973162]
0x2020ff6:                          682c000000  push 44
0x2020ffb:                          e9d0ffffff  jmp 0x02020fd0
0x2021000:                        ff2522a03c00  jmp qword [rip + 3973154]
.plt.got
0x20212c0:                        ff256a9b3c00  jmp qword [rip + 3971946]
0x20212c6:                                6690  nop
0x20212c8:                        ff25ca9c3c00  jmp qword [rip + 3972298]
0x20212ce:                                6690  nop
0x20212d0:                                  55  push rbp
0x20212d1:                                  53  push rbx
0x20212d2:                      488d3d0d251900  lea rdi,qword [rip + 1647885]
0x20212d9:                          be02000080  mov esi,0x80000002
0x20212de:                            4883ec08  sub rsp,8
0x20212e2:                          e869511400  call 0x02166450

linux/amd64/chown.viv
.plt
0x2001d50:                        ff351ae02000  push qword [rip + 2154522]
0x2001d56:                        ff251ce02000  jmp qword [rip + 2154524]
0x2001d5c:                            0f1f4000  nop dword [rax]
0x2001d60:                        ff251ae02000  jmp qword [rip + 2154522]
0x2001d66:                          6800000000  push 0
0x2001d6b:                          e9e0ffffff  jmp 0x02001d50
0x2001d70:                        ff2512e02000  jmp qword [rip + 2154514]
0x2001d76:                          6801000000  push 1
0x2001d7b:                          e9d0ffffff  jmp 0x02001d50
0x2001d80:                        ff250ae02000  jmp qword [rip + 2154506]
.plt.got
0x2002200:                        ff25cadd2000  jmp qword [rip + 2153930]
0x2002206:                                6690  nop
0x2002208:                        ff25eadd2000  jmp qword [rip + 2153962]
0x200220e:                                6690  nop
0x2002210:                                4157  push r15
0x2002212:                                4156  push r14
0x2002214:                              4531f6  xor r14d,r14d
0x2002217:                                4155  push r13
0x2002219:                                4154  push r12
0x200221b:                        41bdffffffff  mov r13d,0xffffffff

openbsd/ls.amd64.viv

qnx/arm/ksh.viv









sections versus dynamics:
In [79]: x+=1; print '======================================\n' + vws[x].metadata['StorageName']; ed=vws[x].filemeta.values()[0]['ELF_DYNAMICS']; print hex(ed['DT_JMPREL']), hex(ed['DT_PLTRELSZ']), hex(ed['DT_PLTGOT']); es[x].sections
======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/amd64/ls.viv
0x4015f8 0x9d8 0x618fe8
Out[79]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [             .interp] @0x00400238 (     568)  ent/size:        0/      28  align:        1,
 Elf Sec: [       .note.ABI-tag] @0x00400254 (     596)  ent/size:        0/      32  align:        4,
 Elf Sec: [  .note.gnu.build-id] @0x00400274 (     628)  ent/size:        0/      36  align:        4,
 Elf Sec: [           .gnu.hash] @0x00400298 (     664)  ent/size:        0/     104  align:        8,
 Elf Sec: [             .dynsym] @0x00400300 (     768)  ent/size:       24/    2904  align:        8,
 Elf Sec: [             .dynstr] @0x00400e58 (    3672)  ent/size:        0/    1361  align:        1,
 Elf Sec: [        .gnu.version] @0x004013aa (    5034)  ent/size:        2/     242  align:        2,
 Elf Sec: [      .gnu.version_r] @0x004014a0 (    5280)  ent/size:        0/     176  align:        8,
 Elf Sec: [           .rela.dyn] @0x00401550 (    5456)  ent/size:       24/     168  align:        8,
 
 Elf Sec: [           .rela.plt] @0x004015f8 (    5624)  ent/size:       24/    2520  align:        8,
 
 Elf Sec: [               .init] @0x00401fd0 (    8144)  ent/size:        0/      24  align:        4,
 Elf Sec: [                .plt] @0x00401ff0 (    8176)  ent/size:       16/    1696  align:       16,
 Elf Sec: [               .text] @0x00402690 (    9872)  ent/size:        0/   61800  align:       16,
 Elf Sec: [               .fini] @0x004117f8 (   71672)  ent/size:        0/      14  align:        4,
 Elf Sec: [             .rodata] @0x00411820 (   71712)  ent/size:        0/   20339  align:       32,
 Elf Sec: [       .eh_frame_hdr] @0x00416794 (   92052)  ent/size:        0/    1732  align:        4,
 Elf Sec: [           .eh_frame] @0x00416e58 (   93784)  ent/size:        0/    7988  align:        8,
 Elf Sec: [              .ctors] @0x00618df0 (  101872)  ent/size:        0/      16  align:        8,
 Elf Sec: [              .dtors] @0x00618e00 (  101888)  ent/size:        0/      16  align:        8,
 Elf Sec: [                .jcr] @0x00618e10 (  101904)  ent/size:        0/       8  align:        8,
 Elf Sec: [            .dynamic] @0x00618e18 (  101912)  ent/size:       16/     448  align:        8,
 Elf Sec: [                .got] @0x00618fd8 (  102360)  ent/size:        8/      16  align:        8,
 
 Elf Sec: [            .got.plt] @0x00618fe8 (  102376)  ent/size:        8/     864  align:        8,
 
 Elf Sec: [               .data] @0x00619360 (  103264)  ent/size:        0/     528  align:       32,
 Elf Sec: [                .bss] @0x00619580 (  103792)  ent/size:        0/    3424  align:       32,
 Elf Sec: [      .gnu_debuglink] @0x00000000 (  103792)  ent/size:        0/       8  align:        1,
 Elf Sec: [           .shstrtab] @0x00000000 (  103800)  ent/size:        0/     244  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/amd64/chown.viv
0x1648 0x6f0 0x20fd68
Out[80]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [             .interp] @0x00000238 (     568)  ent/size:        0/      28  align:        1,
 Elf Sec: [       .note.ABI-tag] @0x00000254 (     596)  ent/size:        0/      32  align:        4,
 Elf Sec: [  .note.gnu.build-id] @0x00000274 (     628)  ent/size:        0/      36  align:        4,
 Elf Sec: [           .gnu.hash] @0x00000298 (     664)  ent/size:        0/     184  align:        8,
 Elf Sec: [             .dynsym] @0x00000350 (     848)  ent/size:       24/    2448  align:        8,
 Elf Sec: [             .dynstr] @0x00000ce0 (    3296)  ent/size:        0/    1108  align:        1,
 Elf Sec: [        .gnu.version] @0x00001134 (    4404)  ent/size:        2/     204  align:        2,
 Elf Sec: [      .gnu.version_r] @0x00001200 (    4608)  ent/size:        0/     112  align:        8,
 Elf Sec: [           .rela.dyn] @0x00001270 (    4720)  ent/size:       24/     984  align:        8,
 
 Elf Sec: [           .rela.plt] @0x00001648 (    5704)  ent/size:       24/    1776  align:        8,
 
 Elf Sec: [               .init] @0x00001d38 (    7480)  ent/size:        0/      23  align:        4,
 Elf Sec: [                .plt] @0x00001d50 (    7504)  ent/size:       16/    1200  align:       16,
 Elf Sec: [            .plt.got] @0x00002200 (    8704)  ent/size:        8/      16  align:        8,
 Elf Sec: [               .text] @0x00002210 (    8720)  ent/size:        0/   36473  align:       16,
 Elf Sec: [               .fini] @0x0000b08c (   45196)  ent/size:        0/       9  align:        4,
 Elf Sec: [             .rodata] @0x0000b0a0 (   45216)  ent/size:        0/    7114  align:       32,
 Elf Sec: [       .eh_frame_hdr] @0x0000cc6c (   52332)  ent/size:        0/    1196  align:        4,
 Elf Sec: [           .eh_frame] @0x0000d118 (   53528)  ent/size:        0/    6216  align:        8,
 Elf Sec: [         .init_array] @0x0020f950 (   63824)  ent/size:        8/       8  align:        8,
 Elf Sec: [         .fini_array] @0x0020f958 (   63832)  ent/size:        8/       8  align:        8,
 Elf Sec: [        .data.rel.ro] @0x0020f960 (   63840)  ent/size:        0/     536  align:       32,
 Elf Sec: [            .dynamic] @0x0020fb78 (   64376)  ent/size:       16/     496  align:        8,
 
 Elf Sec: [                .got] @0x0020fd68 (   64872)  ent/size:        8/     664  align:        8,
 
 Elf Sec: [               .data] @0x00210000 (   65536)  ent/size:        0/     128  align:       32,
 Elf Sec: [                .bss] @0x00210080 (   65664)  ent/size:        0/     456  align:       32,
 Elf Sec: [      .gnu_debuglink] @0x00000000 (   65664)  ent/size:        0/      52  align:        4,
 Elf Sec: [           .shstrtab] @0x00000000 (   65716)  ent/size:        0/     257  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/amd64/libc-2.27.so.viv
0x20b78 0x450 0x3eb000
Out[81]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [  .note.gnu.build-id] @0x00000270 (     624)  ent/size:        0/      36  align:        4,
 Elf Sec: [       .note.ABI-tag] @0x00000294 (     660)  ent/size:        0/      32  align:        4,
 Elf Sec: [           .gnu.hash] @0x000002b8 (     696)  ent/size:        0/   15408  align:        8,
 Elf Sec: [             .dynsym] @0x00003ee8 (   16104)  ent/size:       24/   56040  align:        8,
 Elf Sec: [             .dynstr] @0x000119d0 (   72144)  ent/size:        0/   24286  align:        1,
 Elf Sec: [        .gnu.version] @0x000178ae (   96430)  ent/size:        2/    4670  align:        2,
 Elf Sec: [      .gnu.version_d] @0x00018af0 (  101104)  ent/size:        0/    1028  align:        8,
 Elf Sec: [      .gnu.version_r] @0x00018ef8 (  102136)  ent/size:        0/      48  align:        8,
 Elf Sec: [           .rela.dyn] @0x00018f28 (  102184)  ent/size:       24/   31824  align:        8,
 
 Elf Sec: [           .rela.plt] @0x00020b78 (  134008)  ent/size:       24/    1104  align:        8,
 
 Elf Sec: [                .plt] @0x00020fd0 (  135120)  ent/size:       16/     752  align:       16,
 Elf Sec: [            .plt.got] @0x000212c0 (  135872)  ent/size:        8/      16  align:        8,
 Elf Sec: [               .text] @0x000212d0 (  135888)  ent/size:        0/ 1542508  align:       16,
 Elf Sec: [   __libc_freeres_fn] @0x00199c40 ( 1678400)  ent/size:        0/    6271  align:       16,
 Elf Sec: [__libc_thread_freeres_fn] @0x0019b4c0 ( 1684672)  ent/size:        0/    5058  align:       16,
 Elf Sec: [             .rodata] @0x0019c8a0 ( 1689760)  ent/size:        0/  136360  align:       32,
 Elf Sec: [       .stapsdt.base] @0x001bdd48 ( 1826120)  ent/size:        0/       1  align:        1,
 Elf Sec: [             .interp] @0x001bdd50 ( 1826128)  ent/size:        0/      28  align:       16,
 Elf Sec: [       .eh_frame_hdr] @0x001bdd6c ( 1826156)  ent/size:        0/   23004  align:        4,
 Elf Sec: [           .eh_frame] @0x001c3748 ( 1849160)  ent/size:        0/  129624  align:        8,
 Elf Sec: [   .gcc_except_table] @0x001e31a0 ( 1978784)  ent/size:        0/    1169  align:        1,
 Elf Sec: [               .hash] @0x001e3638 ( 1979960)  ent/size:        4/   13416  align:        8,
 Elf Sec: [              .tdata] @0x003e7620 ( 1996320)  ent/size:        0/      16  align:        8,
 Elf Sec: [               .tbss] @0x003e7630 ( 1996336)  ent/size:        0/     128  align:        8,
 Elf Sec: [         .init_array] @0x003e7630 ( 1996336)  ent/size:        8/       8  align:        8,
 Elf Sec: [   __libc_subfreeres] @0x003e7638 ( 1996344)  ent/size:        0/     256  align:        8,
 Elf Sec: [       __libc_atexit] @0x003e7738 ( 1996600)  ent/size:        0/       8  align:        8,
 Elf Sec: [__libc_thread_subfreeres] @0x003e7740 ( 1996608)  ent/size:        0/      32  align:        8,
 Elf Sec: [   __libc_IO_vtables] @0x003e7760 ( 1996640)  ent/size:        0/    3432  align:       32,
 Elf Sec: [        .data.rel.ro] @0x003e84e0 ( 2000096)  ent/size:        0/    9888  align:       32,
 Elf Sec: [            .dynamic] @0x003eab80 ( 2009984)  ent/size:       16/     480  align:        8,
 Elf Sec: [                .got] @0x003ead60 ( 2010464)  ent/size:        8/     656  align:        8,
 
 Elf Sec: [            .got.plt] @0x003eb000 ( 2011136)  ent/size:        8/     392  align:        8,
 
 Elf Sec: [               .data] @0x003eb1a0 ( 2011552)  ent/size:        0/    5824  align:       32,
 Elf Sec: [                .bss] @0x003ec860 ( 2017376)  ent/size:        0/   17024  align:       32,
 Elf Sec: [       .note.stapsdt] @0x00000000 ( 2017376)  ent/size:        0/    4896  align:        4,
 Elf Sec: [.gnu.warning.sigstack] @0x00000000 ( 2022272)  ent/size:        0/      77  align:       32,
 Elf Sec: [.gnu.warning.sigreturn] @0x00000000 ( 2022368)  ent/size:        0/      50  align:       32,
 Elf Sec: [.gnu.warning.siggetmask] @0x00000000 ( 2022432)  ent/size:        0/      57  align:       32,
 Elf Sec: [ .gnu.warning.tmpnam] @0x00000000 ( 2022496)  ent/size:        0/      55  align:       32,
 Elf Sec: [.gnu.warning.tmpnam_r] @0x00000000 ( 2022560)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.tempnam] @0x00000000 ( 2022624)  ent/size:        0/      56  align:       32,
 Elf Sec: [.gnu.warning.sys_errlist] @0x00000000 ( 2022688)  ent/size:        0/      68  align:       32,
 Elf Sec: [.gnu.warning.sys_nerr] @0x00000000 ( 2022784)  ent/size:        0/      65  align:       32,
 Elf Sec: [   .gnu.warning.gets] @0x00000000 ( 2022880)  ent/size:        0/      57  align:       32,
 Elf Sec: [  .gnu.warning.getpw] @0x00000000 ( 2022944)  ent/size:        0/      58  align:       32,
 Elf Sec: [.gnu.warning.re_max_failures] @0x00000000 ( 2023008)  ent/size:        0/      61  align:       32,
 Elf Sec: [ .gnu.warning.lchmod] @0x00000000 ( 2023072)  ent/size:        0/      47  align:       32,
 Elf Sec: [ .gnu.warning.llseek] @0x00000000 ( 2023136)  ent/size:        0/      63  align:       32,
 Elf Sec: [  .gnu.warning.getwd] @0x00000000 ( 2023200)  ent/size:        0/     122  align:       32,
 Elf Sec: [   .gnu.warning.sstk] @0x00000000 ( 2023328)  ent/size:        0/      45  align:       32,
 Elf Sec: [ .gnu.warning.revoke] @0x00000000 ( 2023392)  ent/size:        0/      47  align:       32,
 Elf Sec: [ .gnu.warning.mktemp] @0x00000000 ( 2023456)  ent/size:        0/      68  align:       32,
 Elf Sec: [   .gnu.warning.gtty] @0x00000000 ( 2023552)  ent/size:        0/      45  align:       32,
 Elf Sec: [   .gnu.warning.stty] @0x00000000 ( 2023616)  ent/size:        0/      45  align:       32,
 Elf Sec: [.gnu.warning.chflags] @0x00000000 ( 2023680)  ent/size:        0/      48  align:       32,
 Elf Sec: [.gnu.warning.fchflags] @0x00000000 ( 2023744)  ent/size:        0/      49  align:       32,
 Elf Sec: [.gnu.warning.__compat_bdflush] @0x00000000 ( 2023808)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.__gets_chk] @0x00000000 ( 2023872)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_space] @0x00000000 ( 2023936)  ent/size:        0/      60  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_init] @0x00000000 ( 2024000)  ent/size:        0/      59  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_append] @0x00000000 ( 2024064)  ent/size:        0/      61  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_alloc] @0x00000000 ( 2024128)  ent/size:        0/      60  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_next] @0x00000000 ( 2024192)  ent/size:        0/      59  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_find] @0x00000000 ( 2024256)  ent/size:        0/      59  align:       32,
 Elf Sec: [ .gnu.warning.getmsg] @0x00000000 ( 2024320)  ent/size:        0/      47  align:       32,
 Elf Sec: [ .gnu.warning.putmsg] @0x00000000 ( 2024384)  ent/size:        0/      47  align:       32,
 Elf Sec: [.gnu.warning.fattach] @0x00000000 ( 2024448)  ent/size:        0/      48  align:       32,
 Elf Sec: [.gnu.warning.fdetach] @0x00000000 ( 2024512)  ent/size:        0/      48  align:       32,
 Elf Sec: [.gnu.warning.setlogin] @0x00000000 ( 2024576)  ent/size:        0/      49  align:       32,
 Elf Sec: [      .gnu_debuglink] @0x00000000 ( 2024628)  ent/size:        0/      20  align:        4,
 Elf Sec: [           .shstrtab] @0x00000000 ( 2024648)  ent/size:        0/    1223  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/amd64/libstdc++.so.6.0.25.viv
0x83788 0x4cf8 0x382000
Out[82]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [  .note.gnu.build-id] @0x00000200 (     512)  ent/size:        0/      36  align:        4,
 Elf Sec: [           .gnu.hash] @0x00000228 (     552)  ent/size:        0/   33808  align:        8,
 Elf Sec: [             .dynsym] @0x00008638 (   34360)  ent/size:       24/  132552  align:        8,
 Elf Sec: [             .dynstr] @0x00028c00 (  166912)  ent/size:        0/  266143  align:        1,
 Elf Sec: [        .gnu.version] @0x00069ba0 (  433056)  ent/size:        2/   11046  align:        2,
 Elf Sec: [      .gnu.version_d] @0x0006c6c8 (  444104)  ent/size:        0/    1292  align:        8,
 Elf Sec: [      .gnu.version_r] @0x0006cbd8 (  445400)  ent/size:        0/     272  align:        8,
 Elf Sec: [           .rela.dyn] @0x0006cce8 (  445672)  ent/size:       24/   92832  align:        8,
 
 Elf Sec: [           .rela.plt] @0x00083788 (  538504)  ent/size:       24/   19704  align:        8,
 
 Elf Sec: [               .init] @0x00088480 (  558208)  ent/size:        0/      23  align:        4,
 Elf Sec: [                .plt] @0x000884a0 (  558240)  ent/size:       16/   13152  align:       16,
 Elf Sec: [            .plt.got] @0x0008b800 (  571392)  ent/size:        0/     184  align:        8,
 Elf Sec: [               .text] @0x0008b8c0 (  571584)  ent/size:        0/  719065  align:       16,
 Elf Sec: [               .fini] @0x0013b19c ( 1290652)  ent/size:        0/       9  align:        4,
 Elf Sec: [             .rodata] @0x0013b1c0 ( 1290688)  ent/size:        0/   33432  align:       32,
 Elf Sec: [       .stapsdt.base] @0x00143458 ( 1324120)  ent/size:        0/       1  align:        1,
 Elf Sec: [       .eh_frame_hdr] @0x0014345c ( 1324124)  ent/size:        0/   30524  align:        4,
 Elf Sec: [           .eh_frame] @0x0014ab98 ( 1354648)  ent/size:        0/  160388  align:        8,
 Elf Sec: [   .gcc_except_table] @0x00171e1c ( 1515036)  ent/size:        0/   25478  align:        4,
 Elf Sec: [               .tbss] @0x00378338 ( 1540920)  ent/size:        0/      32  align:        8,
 Elf Sec: [         .init_array] @0x00378338 ( 1540920)  ent/size:        8/      88  align:        8,
 Elf Sec: [         .fini_array] @0x00378390 ( 1541008)  ent/size:        8/       8  align:        8,
 Elf Sec: [                .jcr] @0x00378398 ( 1541016)  ent/size:        0/       8  align:        8,
 Elf Sec: [        .data.rel.ro] @0x003783a0 ( 1541024)  ent/size:        0/   35240  align:       32,
 Elf Sec: [            .dynamic] @0x00380d48 ( 1576264)  ent/size:       16/     544  align:        8,
 Elf Sec: [                .got] @0x00380f68 ( 1576808)  ent/size:        8/    4248  align:        8,

 Elf Sec: [            .got.plt] @0x00382000 ( 1581056)  ent/size:        8/    6592  align:        8,
 
 Elf Sec: [               .data] @0x003839c0 ( 1587648)  ent/size:        0/     376  align:       32,
 Elf Sec: [                .bss] @0x00383b40 ( 1588024)  ent/size:        0/   14112  align:       32,
 Elf Sec: [       .note.stapsdt] @0x00000000 ( 1588024)  ent/size:        0/     236  align:        4,
 Elf Sec: [      .gnu_debuglink] @0x00000000 ( 1588260)  ent/size:        0/      52  align:        1,
 Elf Sec: [           .shstrtab] @0x00000000 ( 1588312)  ent/size:        0/     316  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/amd64/static64.llvm.elf.viv
---------------------------------------------------------------------------
KeyError                                  Traceback (most recent call last)
/home/atlas/bin/vw in <module>()
----> 1 x+=1; print '======================================\n' + vws[x].metadata['StorageName']; ed=vws[x].filemeta.values()[0]['ELF_DYNAMICS']; print hex(ed['DT_JMPREL']), hex(ed['DT_PLTRELSZ']), hex(ed['DT_PLTGOT']); es[x].sections

KeyError: 'DT_JMPREL'
### Static files don't do Dynamics

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/i386/static32.llvm.elf.viv
---------------------------------------------------------------------------
KeyError                                  Traceback (most recent call last)
/home/atlas/bin/vw in <module>()
----> 1 x+=1; print '======================================\n' + vws[x].metadata['StorageName']; ed=vws[x].filemeta.values()[0]['ELF_DYNAMICS']; print hex(ed['DT_JMPREL']), hex(ed['DT_PLTRELSZ']), hex(ed['DT_PLTGOT']); es[x].sections

KeyError: 'DT_JMPREL'
### Static files don't do Dynamics

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/i386/libc-2.13.so.viv
0x16b38 0x40 0x158ff4
Out[85]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [  .note.gnu.build-id] @0x00000174 (     372)  ent/size:        0/      36  align:        4,
 Elf Sec: [       .note.ABI-tag] @0x00000198 (     408)  ent/size:        0/      32  align:        4,
 Elf Sec: [           .gnu.hash] @0x000001b8 (     440)  ent/size:        4/   15416  align:        4,
 Elf Sec: [             .dynsym] @0x00003df0 (   15856)  ent/size:       16/   37376  align:        4,
 Elf Sec: [             .dynstr] @0x0000cff0 (   53232)  ent/size:        0/   23245  align:        1,
 Elf Sec: [        .gnu.version] @0x00012abe (   76478)  ent/size:        2/    4672  align:        2,
 Elf Sec: [      .gnu.version_d] @0x00013d00 (   81152)  ent/size:        0/     984  align:        4,
 Elf Sec: [      .gnu.version_r] @0x000140d8 (   82136)  ent/size:        0/      64  align:        4,
 Elf Sec: [            .rel.dyn] @0x00014118 (   82200)  ent/size:        8/   10784  align:        4,
 
 Elf Sec: [            .rel.plt] @0x00016b38 (   92984)  ent/size:        8/      64  align:        4,
 
 Elf Sec: [                .plt] @0x00016b78 (   93048)  ent/size:        4/     144  align:        4,
 Elf Sec: [               .text] @0x00016c10 (   93200)  ent/size:        0/ 1083604  align:       16,
 Elf Sec: [   __libc_freeres_fn] @0x0011f4f0 ( 1176816)  ent/size:        0/    4040  align:       16,
 Elf Sec: [__libc_thread_freeres_fn] @0x001204c0 ( 1180864)  ent/size:        0/     386  align:       16,
 Elf Sec: [             .rodata] @0x00120660 ( 1181280)  ent/size:        0/  112648  align:       32,
 Elf Sec: [             .interp] @0x0013be68 ( 1293928)  ent/size:        0/      19  align:        1,
 Elf Sec: [       .eh_frame_hdr] @0x0013be7c ( 1293948)  ent/size:        0/   13116  align:        4,
 Elf Sec: [           .eh_frame] @0x0013f1b8 ( 1307064)  ent/size:        0/   78516  align:        4,
 Elf Sec: [   .gcc_except_table] @0x0015246c ( 1385580)  ent/size:        0/    1473  align:        1,
 Elf Sec: [               .hash] @0x00152a30 ( 1387056)  ent/size:        4/   13444  align:        4,
 Elf Sec: [              .tdata] @0x001571c8 ( 1401288)  ent/size:        0/       8  align:        4,
 Elf Sec: [               .tbss] @0x001571d0 ( 1401296)  ent/size:        0/      56  align:        4,
 Elf Sec: [         .fini_array] @0x001571d0 ( 1401296)  ent/size:        0/       4  align:        4,
 Elf Sec: [              .ctors] @0x001571d4 ( 1401300)  ent/size:        0/      20  align:        4,
 Elf Sec: [              .dtors] @0x001571e8 ( 1401320)  ent/size:        0/       8  align:        4,
 Elf Sec: [   __libc_subfreeres] @0x001571f0 ( 1401328)  ent/size:        0/     112  align:        4,
 Elf Sec: [       __libc_atexit] @0x00157260 ( 1401440)  ent/size:        0/       4  align:        4,
 Elf Sec: [__libc_thread_subfreeres] @0x00157264 ( 1401444)  ent/size:        0/      12  align:        4,
 Elf Sec: [        .data.rel.ro] @0x00157280 ( 1401472)  ent/size:        0/    6908  align:       32,
 Elf Sec: [            .dynamic] @0x00158d7c ( 1408380)  ent/size:        8/     240  align:        4,
 Elf Sec: [                .got] @0x00158e6c ( 1408620)  ent/size:        4/     372  align:        4,
 
 Elf Sec: [            .got.plt] @0x00158ff4 ( 1409012)  ent/size:        4/      44  align:        4,
 
 Elf Sec: [               .data] @0x00159020 ( 1409056)  ent/size:        0/    2428  align:       32,
 Elf Sec: [                .bss] @0x001599a0 ( 1411484)  ent/size:        0/   12392  align:       32,
 Elf Sec: [.gnu.warning.sigstack] @0x00000000 ( 1411488)  ent/size:        0/      77  align:       32,
 Elf Sec: [.gnu.warning.sigreturn] @0x00000000 ( 1411584)  ent/size:        0/      59  align:       32,
 Elf Sec: [.gnu.warning.siggetmask] @0x00000000 ( 1411648)  ent/size:        0/      57  align:       32,
 Elf Sec: [ .gnu.warning.tmpnam] @0x00000000 ( 1411712)  ent/size:        0/      55  align:       32,
 Elf Sec: [.gnu.warning.tmpnam_r] @0x00000000 ( 1411776)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.tempnam] @0x00000000 ( 1411840)  ent/size:        0/      56  align:       32,
 Elf Sec: [.gnu.warning.sys_errlist] @0x00000000 ( 1411904)  ent/size:        0/      68  align:       32,
 Elf Sec: [.gnu.warning.sys_nerr] @0x00000000 ( 1412000)  ent/size:        0/      65  align:       32,
 Elf Sec: [   .gnu.warning.gets] @0x00000000 ( 1412096)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.__memset_zero_constant_len_parameter] @0x00000000 ( 1412153)  ent/size:        0/     184  align:        1,
 Elf Sec: [  .gnu.warning.getpw] @0x00000000 ( 1412352)  ent/size:        0/      58  align:       32,
 Elf Sec: [.gnu.warning.re_max_failures] @0x00000000 ( 1412416)  ent/size:        0/      61  align:       32,
 Elf Sec: [.gnu.warning.setlogin] @0x00000000 ( 1412480)  ent/size:        0/      58  align:       32,
 Elf Sec: [  .gnu.warning.getwd] @0x00000000 ( 1412544)  ent/size:        0/     122  align:       32,
 Elf Sec: [ .gnu.warning.lchmod] @0x00000000 ( 1412672)  ent/size:        0/      56  align:       32,
 Elf Sec: [   .gnu.warning.sstk] @0x00000000 ( 1412736)  ent/size:        0/      54  align:       32,
 Elf Sec: [ .gnu.warning.mktemp] @0x00000000 ( 1412800)  ent/size:        0/      68  align:       32,
 Elf Sec: [   .gnu.warning.gtty] @0x00000000 ( 1412896)  ent/size:        0/      54  align:       32,
 Elf Sec: [   .gnu.warning.stty] @0x00000000 ( 1412960)  ent/size:        0/      54  align:       32,
 Elf Sec: [.gnu.warning.chflags] @0x00000000 ( 1413024)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.fchflags] @0x00000000 ( 1413088)  ent/size:        0/      58  align:       32,
 Elf Sec: [ .gnu.warning.revoke] @0x00000000 ( 1413152)  ent/size:        0/      56  align:       32,
 Elf Sec: [ .gnu.warning.llseek] @0x00000000 ( 1413216)  ent/size:        0/      63  align:       32,
 Elf Sec: [.gnu.warning.__gets_chk] @0x00000000 ( 1413280)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_space] @0x00000000 ( 1413344)  ent/size:        0/      60  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_init] @0x00000000 ( 1413408)  ent/size:        0/      59  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_append] @0x00000000 ( 1413472)  ent/size:        0/      61  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_alloc] @0x00000000 ( 1413536)  ent/size:        0/      60  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_next] @0x00000000 ( 1413600)  ent/size:        0/      59  align:       32,
 Elf Sec: [.gnu.warning.inet6_option_find] @0x00000000 ( 1413664)  ent/size:        0/      59  align:       32,
 Elf Sec: [.gnu.warning.fattach] @0x00000000 ( 1413728)  ent/size:        0/      57  align:       32,
 Elf Sec: [.gnu.warning.fdetach] @0x00000000 ( 1413792)  ent/size:        0/      57  align:       32,
 Elf Sec: [      .gnu_debuglink] @0x00000000 ( 1413849)  ent/size:        0/      20  align:        1,
 Elf Sec: [           .shstrtab] @0x00000000 ( 1413869)  ent/size:        0/    1165  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/i386/libstdc++.so.6.0.25.viv
0x69470 0x1a10 0x182000
Out[86]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [  .note.gnu.build-id] @0x00000134 (     308)  ent/size:        0/      36  align:        4,
 Elf Sec: [           .gnu.hash] @0x00000158 (     344)  ent/size:        4/   33608  align:        4,
 Elf Sec: [             .dynsym] @0x000084a0 (   33952)  ent/size:       16/   87920  align:        4,
 Elf Sec: [             .dynstr] @0x0001dc10 (  121872)  ent/size:        0/  265462  align:        1,
 Elf Sec: [        .gnu.version] @0x0005e906 (  387334)  ent/size:        2/   10990  align:        2,
 Elf Sec: [      .gnu.version_d] @0x000613f4 (  398324)  ent/size:        0/    1436  align:        4,
 Elf Sec: [      .gnu.version_r] @0x00061990 (  399760)  ent/size:        0/     336  align:        4,
 Elf Sec: [            .rel.dyn] @0x00061ae0 (  400096)  ent/size:        8/   31120  align:        4,
 
 Elf Sec: [            .rel.plt] @0x00069470 (  431216)  ent/size:        8/    6672  align:        4,
 
 Elf Sec: [               .init] @0x0006ae80 (  437888)  ent/size:        0/      35  align:        4,
 Elf Sec: [                .plt] @0x0006aeb0 (  437936)  ent/size:        4/   13360  align:       16,
 Elf Sec: [            .plt.got] @0x0006e2e0 (  451296)  ent/size:        8/     184  align:        8,
 Elf Sec: [               .text] @0x0006e3a0 (  451488)  ent/size:        0/  776532  align:       16,
 Elf Sec: [               .fini] @0x0012bcf4 ( 1228020)  ent/size:        0/      20  align:        4,
 Elf Sec: [             .rodata] @0x0012bd20 ( 1228064)  ent/size:        0/   30436  align:       32,
 Elf Sec: [       .stapsdt.base] @0x00133404 ( 1258500)  ent/size:        0/       1  align:        1,
 Elf Sec: [       .eh_frame_hdr] @0x00133408 ( 1258504)  ent/size:        0/   32580  align:        4,
 Elf Sec: [           .eh_frame] @0x0013b34c ( 1291084)  ent/size:        0/  236468  align:        4,
 Elf Sec: [   .gcc_except_table] @0x00174f00 ( 1527552)  ent/size:        0/   27108  align:        4,
 Elf Sec: [               .tbss] @0x0017ceb0 ( 1556144)  ent/size:        0/      16  align:        4,
 Elf Sec: [         .init_array] @0x0017ceb0 ( 1556144)  ent/size:        4/      44  align:        4,
 Elf Sec: [         .fini_array] @0x0017cedc ( 1556188)  ent/size:        4/       4  align:        4,
 Elf Sec: [        .data.rel.ro] @0x0017cee0 ( 1556192)  ent/size:        0/   18276  align:       32,
 Elf Sec: [            .dynamic] @0x00181644 ( 1574468)  ent/size:        8/     272  align:        4,
 Elf Sec: [                .got] @0x00181754 ( 1574740)  ent/size:        4/    2196  align:        4,
 
 Elf Sec: [            .got.plt] @0x00182000 ( 1576960)  ent/size:        4/    3348  align:        4,
 
 Elf Sec: [               .data] @0x00182d20 ( 1580320)  ent/size:        0/     204  align:       32,
 Elf Sec: [                .bss] @0x00182e00 ( 1580524)  ent/size:        0/    8760  align:       32,
 Elf Sec: [       .note.stapsdt] @0x00000000 ( 1580524)  ent/size:        0/     192  align:        4,
 Elf Sec: [      .gnu_debuglink] @0x00000000 ( 1580716)  ent/size:        0/      52  align:        4,
 Elf Sec: [           .shstrtab] @0x00000000 ( 1580768)  ent/size:        0/     309  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//linux/arm/sh.viv
0xaca8 0x928 0x49490
Out[87]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [             .interp] @0x000080d4 (     212)  ent/size:        0/      19  align:        1,
 Elf Sec: [               .hash] @0x000080e8 (     232)  ent/size:        4/    2392  align:        4,
 Elf Sec: [             .dynsym] @0x00008a40 (    2624)  ent/size:       16/    5328  align:        4,
 Elf Sec: [             .dynstr] @0x00009f10 (    7952)  ent/size:        0/    2692  align:        1,
 Elf Sec: [        .gnu.version] @0x0000a994 (   10644)  ent/size:        2/     666  align:        2,
 Elf Sec: [      .gnu.version_r] @0x0000ac30 (   11312)  ent/size:        0/      64  align:        4,
 Elf Sec: [            .rel.dyn] @0x0000ac70 (   11376)  ent/size:        8/      56  align:        4,
 
 Elf Sec: [            .rel.plt] @0x0000aca8 (   11432)  ent/size:        8/    2344  align:        4,
 
 Elf Sec: [               .init] @0x0000b5d0 (   13776)  ent/size:        0/      20  align:        4,
 Elf Sec: [                .plt] @0x0000b5e4 (   13796)  ent/size:        4/    3536  align:        4,
 Elf Sec: [               .text] @0x0000c3b8 (   17336)  ent/size:        0/  175376  align:        8,
 Elf Sec: [               .fini] @0x000370c8 (  192712)  ent/size:        0/      16  align:        4,
 Elf Sec: [             .rodata] @0x000370d8 (  192728)  ent/size:        0/   41656  align:        4,
 Elf Sec: [           .eh_frame] @0x00041390 (  234384)  ent/size:        0/       4  align:        4,
 Elf Sec: [         .init_array] @0x00049394 (  234388)  ent/size:        0/       4  align:        4,
 Elf Sec: [         .fini_array] @0x00049398 (  234392)  ent/size:        0/       4  align:        4,
 Elf Sec: [                .jcr] @0x0004939c (  234396)  ent/size:        0/       4  align:        4,
 Elf Sec: [            .dynamic] @0x000493a0 (  234400)  ent/size:        8/     240  align:        4,
 
 Elf Sec: [                .got] @0x00049490 (  234640)  ent/size:        4/    1192  align:        4,
 
 Elf Sec: [               .data] @0x00049938 (  235832)  ent/size:        0/    1136  align:        4,
 Elf Sec: [                .bss] @0x00049da8 (  236968)  ent/size:        0/   27032  align:        8,
 Elf Sec: [     .ARM.attributes] @0x00000000 (  236968)  ent/size:        0/      16  align:        1,
 Elf Sec: [           .shstrtab] @0x00000000 (  236984)  ent/size:        0/     193  align:        1]

======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//qnx/arm/ksh.viv
0x102d5c 0x7d8 0x146a10
Out[88]: 
[Elf Sec: [                    ] @0x00000000 (       0)  ent/size:        0/       0  align:        0,
 Elf Sec: [            QNX_info] @0x00000000 (  288368)  ent/size:        0/     147  align:        0,
 Elf Sec: [           .shstrtab] @0x00000000 (  288515)  ent/size:        0/      20  align:        1]
### QNX doesn't "do" Sections.


======================================
/home/atlas/hacking/git/atlas0fd00m/vivtestfiles//openbsd/ls.amd64.viv
---------------------------------------------------------------------------
KeyError                                  Traceback (most recent call last)
/home/atlas/bin/vw in <module>()
----> 1 x+=1; print '======================================\n' + vws[x].metadata['StorageName']; ed=vws[x].filemeta.values()[0]['ELF_DYNAMICS']; print hex(ed['DT_JMPREL']), hex(ed['DT_PLTRELSZ']), hex(ed['DT_PLTGOT']); es[x].sections

KeyError: 'DT_JMPREL'
### OpenBSD doesn't "do" Dynamics.


'''



import base64
import logging

import msgpack

logger = logging.getLogger(__name__)

loadargs = {'use_list': False, 'raw': False, 'max_buffer_size': 2**32 -1}
if msgpack.version < (1, 0, 0):
    loadargs['encoding'] = 'utf-8'
else:
    loadargs['strict_map_key'] = False

VSIG = b'MSGVIV'.ljust(8, b'\x00')


def vivEventsAppendFile(filename, events):
    with open(filename, 'ab') as f:
        for event in events:
            if event[0] == 20:
                mape = base64.b64encode(event[1][3])
                event = (event[0], (event[1][0], event[1][1], event[1][2], mape))
            msgpack.pack(event, f, use_bin_type=False)


def saveWorkspaceChanges(vw, filename):
    events = vw.exportWorkspaceChanges()
    vivEventsAppendFile(filename, events)


def vivEventsToFile(filename, events):
    with open(filename, 'wb') as f:
        msgpack.pack(VSIG, f, use_bin_type=False)
        for event in events:
            if event[0] == 20:
                mape = base64.b64encode(event[1][3])
                event = (event[0], (event[1][0], event[1][1], event[1][2], mape, event[1][4]))
            msgpack.pack(event, f, use_bin_type=False)


def saveWorkspace(vw, filename):
    events = vw.exportWorkspace()
    vivEventsToFile(filename, events)


def vivEventsFromFile(filename):
    events = []
    with open(filename, 'rb') as f:
        unpacker = msgpack.Unpacker(f, **loadargs)
        siggy = next(unpacker)
        if siggy.encode('utf-8') != VSIG:
            logger.warning('Invalid file signature of %s', str(siggy))
            return
        for event in unpacker:
            if event[0] == 20:
                mape = base64.b64decode(event[1][3])
                if len(event[1]) == 5:
                    event = (event[0], (event[1][0], event[1][1], event[1][2], mape, event[1][4]))
                else:
                    # DEPRECATED:  for loading older MPFILEs
                    event = (event[0], (event[1][0], event[1][1], event[1][2], mape))
            events.append(event)
    return events


def loadWorkspace(vw, filename):
    events = vivEventsFromFile(filename)
    vw.importWorkspace(events)




"""
Renasas H8 module.
MSB
8-bit
16-bit fixed instruction size (except when EA field, then 32-bit)
"""

import envi


class H8Module(envi.ArchitectureModule):

    def __init__(self, name='h8'):
        envi.ArchitectureModule.__init__(self, name, maxinst=4)
        import envi.archs.h8.disasm as h8_disasm
        self._arch_reg = self.archGetRegCtx()
        self._arch_dis = h8_disasm.H8Disasm()
        self._arch_memreg = {}
        self._arch_memreg.update(memreg_h8s)

    def archGetRegCtx(self):
        import envi.archs.h8.regs as h8_regs
        return h8_regs.H8RegisterContext()

    def archGetBreakInstr(self):
        # TODO: Mop this up. Make a more specific exception or encode the complex instruction
        raise Exception("weird... what are you trying to do here?  h8 has a complex breakpoint instruction")

    def archGetNopInstr(self):
        return b'\x00\x00'

    def getPointerSize(self):
        return 4    # 24 bits encoded in the 3 LSB of 32bit.  addresses must be even

    def pointerString(self, va):
        return "0x%.8x" % va

    def archParseOpcode(self, bytes, offset=0, va=0, extra=None):
        """
        Parse a sequence of bytes out into an envi.Opcode instance.
        """
        return self._arch_dis.disasm(bytes, offset, va)

    def getEmulator(self):
        import envi.archs.h8.emu as h8_emu
        return h8_emu.H8Emulator()

# https://www.renesas.com/us/en/doc/products/mpumcu/001/rej09b0403_h8s2472_2462hm.pdf?key=69aea339d84b503e86d4d19924e46b8c
# see page 1072 onwards for mem -> reg mappings
memreg_h8s = {  # from rej06b0824_h8sap.pdf (h8s/2400).  uncommented entries from IDA H9S/Advanced
        0xffff84: 'SBYCR',     # Standby Control Register
        0xffff86: 'MSTPCRH',
        0xffff87: 'MSTPCRL',
        0xffff89: 'BRR_1',
        0xffff8a: 'SCR_1',
        0xffff8b: 'TDR_1',
        0xffffa0: 'DADRAH',
        0xffffa1: 'BRR_2',
        0xffffa2: 'SCR_2',
        0xffffa3: 'TDR_2',
        0xffffa4: 'SSR_2',
        0xffffa5: 'RDR_2',
        0xffffa8: 'TCNT_0',    # Timer Counter_0(write)  - 16bits
        0xffffa9: 'TCNT_0_byte1',
        0xffffaa: 'PAODR',     # Port A Output Data Register
        0xffffab: 'PA0DDR',    # Port A Data Direction Register
        0xffffb5: 'P4ODDR',    # Port 4 Data Direction Register
        0xffffb7: 'P4ODR',     # Port 4 Data Register
        0xffffbb: 'P6DR',
        0xffffbe: 'P8DDR',
        0xffffbf: 'P8DR',
        0xffffc0: 'P9DDR',
        0xffffc1: 'P9DR',
        0xffffc2: 'IER',
        0xffffc3: 'STCR',
        0xffffc4: 'SYSCR',     # System Control Register
        0xffffc5: 'MDCR',      # Mode Control Register
        0xffffc6: 'BCR',
        0xffffc7: 'WSCR',
        0xffffd8: 'SMR',
        0xffffd9: 'BRR',
        0xffffda: 'SCR',
        0xffffdb: 'TDR',
        0xffffdc: 'SSR',
        0xffffdd: 'RDR',
}


DTC = 0xffec00

DTC_FIELDS = \
        [['External pins',   'IRQ0', '16', "H'0420", ' DTCEA7'],
         ['External pins',   'IRQ1', '17', "H'0422", ' DTCEA6'],
         ['External pins',   'IRQ2', '18', "H'0424", ' DTCEA5'],
         ['External pins',   'IRQ3', '19', "H'0426", ' DTCEA4'],
         ['A D converter',   'ADI',  '28', "H'0438", ' DTCEA3'],
         ['FRT',             'ICIA', '48', "H'0460", ' DTCEA2'],
         ['FRT',             'ICIB', '49', "H'0462", ' DTCEA1'],
         ['FRT',             'OCIA', '52', "H'0468", ' DTCEA0'],
         ['FRT',             'OCIB', '53', "H'046A", ' DTCEB7'],
         ['TMR_0',           'CMIA0', '64', "H'0480", ' DTCEB2'],
         ['TMR_0',           'CMIB0', '65', "H'0482", ' DTCEB1'],
         ['TMR_1',           'CMIA1', '68', "H'0488", ' DTCEB0'],
         ['TMR_1',           'CMIB1', '69', "H'048A", ' DTCEC7'],
         ['TMR_Y',           'CMIAY', '72', "H'0490", ' DTCEC6'],
         ['TMR_Y',           'CMIBY', '73', "H'0492", ' DTCEC5'],
         ['XBS',             'IBF1', '76', "H'0498", ' DTCEC4'],
         ['XBS',             'IBF2', '77', "H'049A", ' DTCEC3'],
         ['SCI_0',           'RXI0', '81', "H'04A2", ' DTCEC2'],
         ['SCI_0',           'TXI0', '82', "H'04A4", ' DTCEC1'],
         ['SCI_1',           'RXI1', '85', "H'04AA", ' DTCEC0'],
         ['SCI_1',           'TXI1', '86', "H'04AC", ' DTCED7'],
         ['SCI_2',           'RXI2', '89', "H'04B2", ' DTCED6'],
         ['SCI_2',           'TXI2', '90', "H'04B4", ' DTCED5'],
         ['IIC_0',           'IICI0', '92', "H'04B8", ' DTCED4'],
         ['IIC_1',           'IICI1', '94', "H'04BC", ' DTCED3'],
         ['LPC',             'ERRI', '108', "H'04D8", ' DTCEE3'],
         ['LPC',             'IBFI1', '109', "H'04DA", ' DTCEE2'],
         ['LPC',             'IBFI2', '110', "H'04DC", ' DTCEE1'],
         ['LPC',             'IBFI3', '111', "H'04DE", ' DTCEE0'],
         ['Software',        'DTVECR_0', '0', "H'0400", ''],
         ['Software',        'DTVECR_1', '1', "H'0402", ''],
         ['Software',        'DTVECR_2', '2', "H'0404", ''],
         ['Software',        'DTVECR_3', '3', "H'0406", ''],
         ['Software',        'DTVECR_4', '4', "H'0408", ''],
         ['Software',        'DTVECR_5', '5', "H'040a", ''],
         ['Software',        'DTVECR_6', '6', "H'040c", ''],
         ['Software',        'DTVECR_7', '7', "H'040e", ''],
         ['Software',        'DTVECR_8', '8', "H'0410", ''],
         ['Software',        'DTVECR_9', '9', "H'0412", '']]

dtc_h8s = {DTC + int(off[2:], 16): "%s_%s" % (n1.replace(" ", "_"), n2) for n1, n2, num, off, ln in DTC_FIELDS}



import os
import hashlib
import tempfile
import unittest

import vivisect
import vivisect.storage.basicfile as vsbf
from vivisect.const import *


def add_events(vw):
    # Call some APIs to populate the event list.
    # Coherency is not the objective. Just get some events in there.
    filehash = hashlib.md5(b'testfile').hexdigest()
    vw.addFile('testfile', 0x1000, filehash)
    vw.setMeta('Architecture', 'i386')
    vw.setMeta('Format', 'pe')
    vw.addMemoryMap(0x1000, 7, 'testfile', b'\x00' * 0x9000)
    vw.addLocation(0x2000, 4, 4, tinfo='fakeptr')
    vw.addLocation(0x3000, 16, 6, tinfo='oogieboogie')
    vw.addLocation(0x4000, 3, 5)
    vw.addLocation(0x5000, 3, 5)
    vw.addLocation(0x6000, 3, 5)
    vw.delLocation(0x4000)
    vw.addXref(0x4000, 0x3000, 1)
    vw.addXref(0x5000, 0x3000, 1)
    vw.delXref((0x5000, 0x3000, 1, 0))
    vw.setMeta('foo', 'bar')
    vw.setFileMeta('testfile', 'neato', 'burrito')
    vw.addVaSet('EmptySet', (('va', VASET_ADDRESS),))
    vw.delVaSet('FuncWrappers')
    vw.setComment(0x2000, 'test comment')
    vw.addExport(0x7000, EXP_FUNCTION, 'kernel32.YoThisExportFake', 'testfile')


class StorageTests(unittest.TestCase):

    def setUp(self):
        '''
        So on windows, you can't double open a temporary file (results in a fun "Permission Denied"
        exception). So instead, we setup a temporary file here and delete it in tearDown so that
        we don't maintain an open file descriptor to the temporary file
        '''
        self.tmpf = tempfile.NamedTemporaryFile(delete=False)

    def tearDown(self):
        self.tmpf.close()
        os.unlink(self.tmpf.name)

    def test_msgpack_idempotent(self):
        # test that what we put in, we can get out
        vw = vivisect.VivWorkspace()
        vw.setMeta('StorageName', self.tmpf.name)
        vw.setMeta('StorageModule', 'vivisect.storage.mpfile')
        add_events(vw)
        vw.saveWorkspace()
        self.tmpf.flush()

        ovw = vivisect.VivWorkspace()
        ovw.setMeta('StorageModule', 'vivisect.storage.mpfile')
        # So this is a bit naughty, but just the act of creating a workspace
        # induces some events in to the workspace. Nothing crazy, just some va sets
        # so delete those so we can have a clean comparison
        ovw._event_list = []
        ovw.loadWorkspace(self.tmpf.name)

        old = list(vw.exportWorkspace())
        new = list(ovw.exportWorkspace())
        self.assertEqual(len(old), 40)
        self.assertEqual(len(new), 41)  # the last event is a setMeta made by loadWorkspace
        self.assertEqual(new[-1], (VWE_SETMETA, ('StorageName', self.tmpf.name)))
        for idx in range(len(old)):
            self.assertEqual(old[idx], new[idx])

    def test_msgpack_to_basicfile(self):
        # make sure we're on par with what the OG storage mechanism can do
        mpfile = tempfile.NamedTemporaryFile(delete=False)
        basicfile = tempfile.NamedTemporaryFile(delete=False)

        try:
            ogvw = vivisect.VivWorkspace()
            add_events(ogvw)
            ogvw.setMeta('StorageName', mpfile.name)
            ogvw.setMeta('StorageModule', 'vivisect.storage.mpfile')
            ogvw.saveWorkspace()
            # Get rid of those last two meta sets so that the two new workspaces should be
            # the same save for the last meta set
            ogvw._event_list.pop()
            ogvw._event_list.pop()
            ogvw.setMeta('StorageName', basicfile.name)
            ogvw.setMeta('StorageModule', 'vivisect.storage.basicfile')
            ogvw.saveWorkspace()
            ogevt = list(ogvw.exportWorkspace())

            mvw = vivisect.VivWorkspace()
            mvw.setMeta('StorageModule', 'vivisect.storage.mpfile')
            mvw._event_list = []
            mvw.loadWorkspace(mpfile.name)
            mevt = list(mvw.exportWorkspace())
            self.assertEqual(len(mevt), 41)

            bvw = vivisect.VivWorkspace()
            bvw.setMeta('StorageModule', 'vivisect.storage.basicfile')
            bvw._event_list = []
            bvw.loadWorkspace(basicfile.name)
            bevt = list(bvw.exportWorkspace())
            self.assertEqual(len(bevt), 41)

            # the last three events are specific to the different storage modules
            for idx in range(len(mevt) - 3):
                self.assertEqual(mevt[idx], bevt[idx])
                self.assertEqual(ogevt[idx], bevt[idx])
        finally:
            mpfile.close()
            basicfile.close()
            os.unlink(mpfile.name)
            os.unlink(basicfile.name)

    def test_bad_event(self):
        vw = vivisect.VivWorkspace()
        with self.assertLogs() as logcap:
            vw.importWorkspace([(VWE_MAX + 1, (0xabad1dea, 4, 3, 'nope')),
                                (VWE_ADDFILE, ('VivisectFile', 0x1000, '3bfdad02b9a6522c84e356cf8f69135b'))])
        files = vw.getFiles()
        self.assertIn("IndexError: list index out of range", ''.join(logcap.output))
        self.assertEqual(1, len(files))
        self.assertEqual('VivisectFile', files[0])

    def test_basicfile_header(self):
        basicfile = tempfile.NamedTemporaryFile(delete=False)
        vw = vivisect.VivWorkspace()
        add_events(vw)
        vw.setMeta('StorageName', basicfile.name)
        vw.saveWorkspace(False)

        # test the header was added correctly
        with open(basicfile.name, 'rb') as f:
            header = f.read(8)
            self.assertEqual(header, vsbf.vivsig_cpickle)

        # now add events and make sure we don't add the header unnecessarily
        vw = vivisect.VivWorkspace()
        vw.loadWorkspace(basicfile.name)
        vw.addLocation(0x6100, 3, 5)
        vw.addLocation(0x6105, 3, 5)
        vw.addLocation(0x610a, 3, 5)
        vw.saveWorkspace(False)

        # now load it.  if a VIV\0\0\0\0\0 was added incorrectly, this will blow up
        vw = vivisect.VivWorkspace()
        vw.loadWorkspace(basicfile.name)
        self.assertEqual(vw.getLocation(0x6100), (0x6100, 3, 5, None))
        self.assertEqual(vw.getLocation(0x6105), (0x6105, 3, 5, None))
        self.assertEqual(vw.getLocation(0x610a), (0x610a, 3, 5, None))




import binascii

from vstruct import VStruct
from vstruct.primitives import *

class v_bits(v_number):

    def __init__(self, width):
        v_number.__init__(self)
        self._vs_bitwidth = int(width)

    def vsSetValue(self, value):
        '''
        override base because we are not using the min or max val.
        '''
        self._vs_value = value

    def vsSetBitWidth(self, width):
        self._vs_bitwidth = width

class VBitField(VStruct):
    '''
    A VStruct *like* field which may contain v_bits
    children.  To all VStruct related operations, we are
    a primitive, but we do have internal structure.

    Example:
        see vstruct/defs/swf.py

    NOTE: this object will pad itself to byte aligned bounds
    '''
    def __init__(self):
        VStruct.__init__(self)

    def vsIsPrim(self):
        return True

    def vsAddField(self, name, value):
        if not isinstance(value, v_bits):
            raise Exception('VBitField *must* use v_bits() kids!')
        return VStruct.vsAddField(self, name, value)

    def vsGetPrintInfo(self, offset=0, indent=0, top=True):
        ret = []
        if top:
            ret.append((offset, indent, self._vs_name, self))

        indent += 1
        bitoff = 0
        for fname,field in self.vsGetFields():
            # use vsSetBitWidth(0) to disable fields
            if field._vs_bitwidth == 0:
                continue
            bitname = '%s[%d:%d]' % (fname,bitoff,bitoff + field._vs_bitwidth)
            ret.append( (offset, indent, bitname, field) )
            bitoff += field._vs_bitwidth

        return ret

    def __len__(self):
        bits = sum([ f._vs_bitwidth for (n,f) in self.vsGetFields() ])
        bittobyte,bitoff = divmod(bits,8)
        if bitoff:
            bittobyte += 1
        return bittobyte

    def vsParse(self, bytez, offset=0):
        bitoff = 0

        for fname,field in self.vsGetFields():

            # use vsSetBitWidth(0) to disable fields
            if field._vs_bitwidth == 0:
                continue

            # adjust forward from last fields bits % 8
            startbyte,startbit = divmod(bitoff,8)
            #offset += bittobyte

            endbyte,endbit = divmod(bitoff + field._vs_bitwidth, 8)
            # if we have an endbit remainder, we need to grab
            # an additional byte...
            endround = 0
            endshift = 0
            if endbit:
                endshift = (8-endbit)
                endround = 1

            fieldbytes = bytez[offset + startbyte:offset+endbyte+endround]
            rawint = int(binascii.hexlify(fieldbytes), 16)
            if endshift:
            #if bitshift:
                rawint >>= endshift

            rawint &= (2**field._vs_bitwidth)-1
            field.vsSetValue(rawint)
            bitoff += field._vs_bitwidth

            self._vsFireCallbacks(fname)

        offbytes,offbits = divmod(bitoff,8)
        offset += offbytes
        # mop up any remaining bits int a byte boundary
        if offbits:
            offset += 1

        return offset

    def vsEmit(self):
        valu = 0
        width = 0

        for name,field in self.vsGetFields():
                width += field._vs_bitwidth
                valu = ( valu << field._vs_bitwidth ) | field._vs_value
        bytelen,bitrem = divmod(width,8)
        if bitrem:
            bytelen += 1
            valu <<= ( 8 - bitrem )

        return binascii.unhexlify(('%.' + str(bytelen*2) + 'x') % valu)



import os
import sys
import getpass
import hashlib
import argparse
import binascii

import cobra.auth as c_auth


class ShadowFileAuth(c_auth.CobraAuthenticator):

    '''
    An auth module which uses a simple text file with salted SHA256
    password hashes.  ( this module may be executed as main to produce
    rows for the file )

    File Format:
    # comment
    <username>:<salt>$<saltedhash>
    '''
    def __init__(self, filename):
        c_auth.CobraAuthenticator.__init__(self)
        self.filename = filename
        if not os.path.isfile(filename):
            raise Exception('No Such File: %s' % filename)

    def authCobraUser(self, authinfo):
        user = authinfo.get('user').lower()
        passwd = authinfo.get('passwd')
        userhdr = '%s:' % user
        with open(self.filename, 'r') as fd:
            for line in fd.readlines():
                if not line.startswith(userhdr):
                    continue

                line = line.strip()
                # We are on the correct line
                suser, spasswd = line.split(':')
                salt, pwhash = spasswd.split('$')
                enc = salt + passwd
                if hashlib.sha256(enc.encode('utf-8')).hexdigest() == pwhash:
                    return user

                break

        return None


def setup():
    desc = 'Helper tool for making shadow file rows'
    ap = argparse.ArgumentParser('cobra.auth.shadowfile', description=desc)
    ap.add_argument('user', help='username to create password for')
    ap.add_argument('--passwd', default=getpass.getpass(), help='optional password to hash')

    return ap


def main(argv):
    opts = setup().parse_args(argv)
    user = opts.user.lower()

    salt = binascii.hexlify(os.urandom(8))
    hash = hashlib.sha256(salt + opts.passwd).hexdigest()
    print('%s:%s$%s' % (user, salt, hash))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import envi.archs.h8.const as h8_const
import envi.registers as e_reg

h8_regs = (
    ('er0', 32),
    ('er1', 32),
    ('er2', 32),
    ('er3', 32),
    ('er4', 32),
    ('er5', 32),
    ('er6', 32),
    ('sp', 32),
    ('pc', 24),
    ('ccr', 8),
    ('exr', 8),
)


l = locals()
e_reg.addLocalEnums(l, h8_regs)

REG_CCR_T = 7
REG_CCR_U1 = 6
REG_CCR_H = 5
REG_CCR_U0 = 4
REG_CCR_N = 3
REG_CCR_Z = 2
REG_CCR_V = 1
REG_CCR_C = 0

ccr_fields = [
    REG_CCR_C,
    REG_CCR_V,
    REG_CCR_Z,
    REG_CCR_N,
    REG_CCR_U0,
    REG_CCR_H,
    REG_CCR_U1,
    REG_CCR_T,
]

H8StatMeta = tuple([
    ("N", h8_const.REG_FLAGS, REG_CCR_N, 1, 'Negative'),
    ("Z", h8_const.REG_FLAGS, REG_CCR_Z, 1, 'Zero'),
    ("C", h8_const.REG_FLAGS, REG_CCR_C, 1, 'Carry'),
    ("V", h8_const.REG_FLAGS, REG_CCR_V, 1, 'oVerflow'),
    ("U0", h8_const.REG_FLAGS, REG_CCR_U0, 1, 'User 0'),
    ("U1", h8_const.REG_FLAGS, REG_CCR_U1, 1, 'User 0'),
    ("T", h8_const.REG_FLAGS, REG_CCR_T, 1, 'T'),
    ("H", h8_const.REG_FLAGS, REG_CCR_H, 1, 'Half Carry'),
    ])


H8Meta = [
    ('r0', REG_ER0, 0, 16),
    ('e0', REG_ER0, 16, 16),
    ('r0h', REG_ER0, 8, 8),
    ('r0l', REG_ER0, 0, 8),
    ('r1', REG_ER1, 0, 16),
    ('e1', REG_ER1, 16, 16),
    ('r1h', REG_ER1, 8, 8),
    ('r1l', REG_ER1, 0, 8),
    ('r2', REG_ER2, 0, 16),
    ('e2', REG_ER2, 16, 16),
    ('r2h', REG_ER2, 8, 8),
    ('r2l', REG_ER2, 0, 8),
    ('r3', REG_ER3, 0, 16),
    ('e3', REG_ER3, 16, 16),
    ('r3h', REG_ER3, 8, 8),
    ('r3l', REG_ER3, 0, 8),
    ('r4', REG_ER4, 0, 16),
    ('e4', REG_ER4, 16, 16),
    ('r4h', REG_ER4, 8, 8),
    ('r4l', REG_ER4, 0, 8),
    ('r5', REG_ER5, 0, 16),
    ('e5', REG_ER5, 16, 16),
    ('r5h', REG_ER5, 8, 8),
    ('r5l', REG_ER5, 0, 8),
    ('r6', REG_ER6, 0, 16),
    ('e6', REG_ER6, 16, 16),
    ('r6h', REG_ER6, 8, 8),
    ('r6l', REG_ER6, 0, 8),
    ('r7', REG_SP, 0, 16),
    ('e7', REG_SP, 16, 16),
    ('r7h', REG_SP, 8, 8),
    ('r7l', REG_SP, 0, 8),
    ('er7', REG_SP, 0, 32),
]


def metaFrom8(regidx):
    width = 8
    extended = regidx >> 3
    offset = (width, 0)[extended]

    idx = regidx & 0x7
    idx |= (width << 16)
    idx |= (offset << 24)

    return idx


def metaFrom16(regidx):
    width = 16
    extended = regidx >> 3
    offset = (0, width)[extended]

    idx = regidx & 0x7
    idx |= (width << 16)
    idx |= (offset << 24)

    return idx


converters = (
        None,
        metaFrom8,
        metaFrom16,
        None,
        None,
        )


def convertMeta(regidx, tsize):
    converter = converters[tsize]
    if converter is None:
        return regidx
    return converter(regidx)


e_reg.addLocalStatusMetas(l, H8Meta, H8StatMeta, 'CCR')


class H8RegisterContext(e_reg.RegisterContext):
    def __init__(self):
        e_reg.RegisterContext.__init__(self)
        self.loadRegDef(h8_regs)
        self.loadRegMetas(H8Meta, statmetas=H8StatMeta)
        self.setRegisterIndexes(REG_PC, h8_const.REG_SP, REG_CCR)
